#!/usr/bin/env python3
"""
# random_seed.py
#
# Test fio's random seed options.
#
# - make sure that randseed overrides randrepeat and allrandrepeat
# - make sure that seeds differ across invocations when [all]randrepeat=0 and randseed is not set
# - make sure that seeds are always the same when [all]randrepeat=1 and randseed is not set
#
# USAGE
# see python3 random_seed.py --help
#
# EXAMPLES
# python3 t/random_seed.py
# python3 t/random_seed.py -f ./fio
#
# REQUIREMENTS
# Python 3.6
#
"""
import os
import sys
import time
import locale
import logging
import argparse
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests

class FioRandTest(FioJobCmdTest):
    """fio random seed test."""

    def setup(self, parameters):
        """Setup the test."""

        fio_args = [
            "--debug=random",
            "--name=random_seed",
            "--ioengine=null",
            "--filesize=32k",
            "--rw=randread",
            f"--output={self.filenames['output']}",
        ]
        for opt in ['randseed', 'randrepeat', 'allrandrepeat']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)

    def get_rand_seeds(self):
        """Collect random seeds from --debug=random output."""
        with open(self.filenames['output'], "r",
                  encoding=locale.getpreferredencoding()) as out_file:
            file_data = out_file.read()

            offsets = 0
            for line in file_data.split('\n'):
                if 'random' in line and 'FIO_RAND_NR_OFFS=' in line:
                    tokens = line.split('=')
                    offsets = int(tokens[len(tokens)-1])
                    break

            if offsets == 0:
                pass
                # find an exception to throw

            seed_list = []
            for line in file_data.split('\n'):
                if 'random' not in line:
                    continue
                if 'rand_seeds[' in line:
                    tokens = line.split('=')
                    seed = int(tokens[-1])
                    seed_list.append(seed)
                    # assume that seeds are in order

            return seed_list


class TestRR(FioRandTest):
    """
    Test object for [all]randrepeat. If run for the first time just collect the
    seeds. For later runs make sure the seeds match or do not match those
    previously collected.
    """
    # one set of seeds is for randrepeat=0 and the other is for randrepeat=1
    seeds = { 0: None, 1: None }

    def check_result(self):
        """Check output for allrandrepeat=1."""

        super().check_result()
        if not self.passed:
            return

        opt = 'randrepeat' if 'randrepeat' in self.fio_opts else 'allrandrepeat'
        rr = self.fio_opts[opt]
        rand_seeds = self.get_rand_seeds()

        if not TestRR.seeds[rr]:
            TestRR.seeds[rr] = rand_seeds
            logging.debug("TestRR: saving rand_seeds for [a]rr=%d", rr)
        else:
            if rr:
                if TestRR.seeds[1] != rand_seeds:
                    self.passed = False
                    print(f"TestRR: unexpected seed mismatch for [a]rr={rr}")
                else:
                    logging.debug("TestRR: seeds correctly match for [a]rr=%d", rr)
                if TestRR.seeds[0] == rand_seeds:
                    self.passed = False
                    print("TestRR: seeds unexpectedly match those from system RNG")
            else:
                if TestRR.seeds[0] == rand_seeds:
                    self.passed = False
                    print(f"TestRR: unexpected seed match for [a]rr={rr}")
                else:
                    logging.debug("TestRR: seeds correctly don't match for [a]rr=%d", rr)
                if TestRR.seeds[1] == rand_seeds:
                    self.passed = False
                    print("TestRR: random seeds unexpectedly match those from [a]rr=1")


class TestRS(FioRandTest):
    """
    Test object when randseed=something controls the generated seeds. If run
    for the first time for a given randseed just collect the seeds. For later
    runs with the same seed make sure the seeds are the same as those
    previously collected.
    """
    seeds = {}

    def check_result(self):
        """Check output for randseed=something."""

        super().check_result()
        if not self.passed:
            return

        rand_seeds = self.get_rand_seeds()
        randseed = self.fio_opts['randseed']

        logging.debug("randseed = %s", randseed)

        if randseed not in TestRS.seeds:
            TestRS.seeds[randseed] = rand_seeds
            logging.debug("TestRS: saving rand_seeds")
        else:
            if TestRS.seeds[randseed] != rand_seeds:
                self.passed = False
                print("TestRS: seeds don't match when they should")
            else:
                logging.debug("TestRS: seeds correctly match")

        # Now try to find seeds generated using a different randseed and make
        # sure they *don't* match
        for key, value in TestRS.seeds.items():
            if key != randseed:
                if value == rand_seeds:
                    self.passed = False
                    print("TestRS: randseeds differ but generated seeds match.")
                else:
                    logging.debug("TestRS: randseeds differ and generated seeds also differ.")


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-d', '--debug', help='enable debug output', action='store_true')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    args = parser.parse_args()

    return args


def main():
    """Run tests of fio random seed options"""

    args = parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    artifact_root = args.artifact_root if args.artifact_root else \
        f"random-seed-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    test_list = [
        {
            "test_id": 1,
            "fio_opts": {
                "randrepeat": 0,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 2,
            "fio_opts": {
                "randrepeat": 0,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 3,
            "fio_opts": {
                "randrepeat": 1,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 4,
            "fio_opts": {
                "randrepeat": 1,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 5,
            "fio_opts": {
                "allrandrepeat": 0,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 6,
            "fio_opts": {
                "allrandrepeat": 0,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 7,
            "fio_opts": {
                "allrandrepeat": 1,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 8,
            "fio_opts": {
                "allrandrepeat": 1,
                },
            "test_class": TestRR,
        },
        {
            "test_id": 9,
            "fio_opts": {
                "randrepeat": 0,
                "randseed": "12345",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 10,
            "fio_opts": {
                "randrepeat": 0,
                "randseed": "12345",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 11,
            "fio_opts": {
                "randrepeat": 1,
                "randseed": "12345",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 12,
            "fio_opts": {
                "allrandrepeat": 0,
                "randseed": "12345",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 13,
            "fio_opts": {
                "allrandrepeat": 1,
                "randseed": "12345",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 14,
            "fio_opts": {
                "randrepeat": 0,
                "randseed": "67890",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 15,
            "fio_opts": {
                "randrepeat": 1,
                "randseed": "67890",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 16,
            "fio_opts": {
                "allrandrepeat": 0,
                "randseed": "67890",
                },
            "test_class": TestRS,
        },
        {
            "test_id": 17,
            "fio_opts": {
                "allrandrepeat": 1,
                "randseed": "67890",
                },
            "test_class": TestRS,
        },
    ]

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'random',
              }

    _, failed, _ = run_fio_tests(test_list, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# log_compression.py
#
# Test log_compression and log_store_compressed. Uses null ioengine.
# Previous bugs have caused output in per I/O log files to be missing
# and/or out of order
#
# Expected result: 8000 log entries, offset starting at 0 and increasing by bs
# Buggy result: Log entries out of order (usually without log_store_compressed)
# and/or missing log entries (usually with log_store_compressed)
#
# USAGE
# python log_compression.py [-f fio-executable]
#
# EXAMPLES
# python t/log_compression.py
# python t/log_compression.py -f ./fio
#
# REQUIREMENTS
# Python 3.5+
#
# ===TEST MATRIX===
#
# With log_compression=10K
# With log_store_compressed=1 and log_compression=10K

import os
import sys
import platform
import argparse
import subprocess


def parse_args():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio',
                        help='path to fio executable (e.g., ./fio)')
    return parser.parse_args()


def run_fio(fio,log_store_compressed):
    fio_args = [
        '--name=job',
        '--ioengine=null',
        '--filesize=1000M',
        '--bs=128K',
        '--rw=write',
        '--iodepth=1',
        '--write_bw_log=test',
        '--per_job_logs=0',
        '--log_offset=1',
        '--log_compression=10K',
        ]
    if log_store_compressed:
        fio_args.append('--log_store_compressed=1')

    subprocess.check_output([fio] + fio_args)

    if log_store_compressed:
        fio_inflate_args = [
            '--inflate-log=test_bw.log.fz'
            ]
        with open('test_bw.from_fz.log','wt') as f:
            subprocess.check_call([fio]+fio_inflate_args,stdout=f)

def check_log_file(log_store_compressed):
    filename = 'test_bw.from_fz.log' if log_store_compressed else 'test_bw.log'
    with open(filename,'rt') as f:
        file_data = f.read()
    log_lines = [x for x in file_data.split('\n') if len(x.strip())!=0]
    log_ios = len(log_lines)

    filesize = 1000*1024*1024
    bs = 128*1024
    ios = filesize//bs
    if log_ios!=ios:
        print('wrong number of ios ({}) in log; should be {}'.format(log_ios,ios))
        return False

    expected_offset = 0
    for line_number,line in enumerate(log_lines):
        log_offset = int(line.split(',')[4])
        if log_offset != expected_offset:
            print('wrong offset ({}) for io number {} in log; should be {}'.format(
                log_offset, line_number, expected_offset))
            return False
        expected_offset += bs
    return True

def main():
    """Entry point for this script."""
    args = parse_args()
    if args.fio:
        fio_path = args.fio
    else:
        fio_path = os.path.join(os.path.dirname(__file__), '../fio')
        if not os.path.exists(fio_path):
            fio_path = 'fio'
    print("fio path is", fio_path)

    passed_count = 0
    failed_count = 0
    for log_store_compressed in [False, True]:
        run_fio(fio_path, log_store_compressed)
        passed = check_log_file(log_store_compressed)
        print('Test with log_store_compressed={} {}'.format(log_store_compressed,
            'PASSED' if passed else 'FAILED'))
        if passed:
            passed_count+=1
        else:
            failed_count+=1

    print('{} tests passed, {} failed'.format(passed_count, failed_count))

    sys.exit(failed_count)

if __name__ == '__main__':
    main()




#!/usr/bin/env python3
"""
fiotestlib.py

This library contains FioTest objects that provide convenient means to run
different sorts of fio tests.

It also contains a test runner that runs an array of dictionary objects
describing fio tests.
"""

import os
import sys
import json
import locale
import logging
import platform
import traceback
import subprocess
from pathlib import Path
from fiotestcommon import get_file, SUCCESS_DEFAULT


class FioTest():
    """Base for all fio tests."""

    def __init__(self, exe_path, success, testnum, artifact_root):
        self.success = success
        self.testnum = testnum
        self.output = {}
        self.passed = True
        self.failure_reason = ''
        self.parameters = None
        self.paths = {
                        'exe': exe_path,
                        'artifacts': artifact_root,
                        'test_dir': os.path.join(artifact_root, \
                                f"{testnum:04d}"),
                        }
        self.filenames = {
                            'cmd': os.path.join(self.paths['test_dir'], \
                                    f"{os.path.basename(self.paths['exe'])}.command"),
                            'stdout': os.path.join(self.paths['test_dir'], \
                                    f"{os.path.basename(self.paths['exe'])}.stdout"),
                            'stderr': os.path.join(self.paths['test_dir'], \
                                    f"{os.path.basename(self.paths['exe'])}.stderr"),
                            'exitcode': os.path.join(self.paths['test_dir'], \
                                    f"{os.path.basename(self.paths['exe'])}.exitcode"),
                            }

    def setup(self, parameters):
        """Setup instance variables for test."""

        self.parameters = parameters
        if not os.path.exists(self.paths['test_dir']):
            os.mkdir(self.paths['test_dir'])

    def run(self):
        """Run the test."""

        raise NotImplementedError()

    def check_result(self):
        """Check test results."""

        raise NotImplementedError()


class FioExeTest(FioTest):
    """Test consists of an executable binary or script"""

    def run(self):
        """Execute the binary or script described by this instance."""

        command = [self.paths['exe']] + self.parameters
        with open(self.filenames['cmd'], "w+",
                  encoding=locale.getpreferredencoding()) as command_file:
            command_file.write(" \\\n ".join(command))

        try:
            with open(self.filenames['stdout'], "w+",
                      encoding=locale.getpreferredencoding()) as stdout_file, \
                open(self.filenames['stderr'], "w+",
                     encoding=locale.getpreferredencoding()) as stderr_file, \
                open(self.filenames['exitcode'], "w+",
                     encoding=locale.getpreferredencoding()) as exitcode_file:
                proc = None
                # Avoid using subprocess.run() here because when a timeout occurs,
                # fio will be stopped with SIGKILL. This does not give fio a
                # chance to clean up and means that child processes may continue
                # running and submitting IO.
                proc = subprocess.Popen(command,
                                        stdout=stdout_file,
                                        stderr=stderr_file,
                                        cwd=self.paths['test_dir'],
                                        universal_newlines=True)
                proc.communicate(timeout=self.success['timeout'])
                exitcode_file.write(f'{proc.returncode}\n')
                logging.debug("Test %d: return code: %d", self.testnum, proc.returncode)
                self.output['proc'] = proc
        except subprocess.TimeoutExpired:
            proc.terminate()
            proc.communicate()
            assert proc.poll()
            self.output['failure'] = 'timeout'
        except Exception:
            if proc:
                if not proc.poll():
                    proc.terminate()
                    proc.communicate()
            self.output['failure'] = 'exception'
            self.output['exc_info'] = sys.exc_info()

    def check_result(self):
        """Check results of test run."""

        if 'proc' not in self.output:
            if self.output['failure'] == 'timeout':
                self.failure_reason = f"{self.failure_reason} timeout,"
            else:
                assert self.output['failure'] == 'exception'
                self.failure_reason = f'{self.failure_reason} exception: ' + \
                f'{self.output["exc_info"][0]}, {self.output["exc_info"][1]}'

            self.passed = False
            return

        if 'zero_return' in self.success:
            if self.success['zero_return']:
                if self.output['proc'].returncode != 0:
                    self.passed = False
                    self.failure_reason = f"{self.failure_reason} non-zero return code,"
            else:
                if self.output['proc'].returncode == 0:
                    self.failure_reason = f"{self.failure_reason} zero return code,"
                    self.passed = False

        stderr_size = os.path.getsize(self.filenames['stderr'])
        if 'stderr_empty' in self.success:
            if self.success['stderr_empty']:
                if stderr_size != 0:
                    self.failure_reason = f"{self.failure_reason} stderr not empty,"
                    self.passed = False
            else:
                if stderr_size == 0:
                    self.failure_reason = f"{self.failure_reason} stderr empty,"
                    self.passed = False


class FioJobFileTest(FioExeTest):
    """Test consists of a fio job with options in a job file."""

    def __init__(self, fio_path, fio_job, success, testnum, artifact_root,
                 fio_pre_job=None, fio_pre_success=None,
                 output_format="normal"):
        """Construct a FioJobFileTest which is a FioExeTest consisting of a
        single fio job file with an optional setup step.

        fio_path:           location of fio executable
        fio_job:            location of fio job file
        success:            Definition of test success
        testnum:            test ID
        artifact_root:      root directory for artifacts
        fio_pre_job:        fio job for preconditioning
        fio_pre_success:    Definition of test success for fio precon job
        output_format:      normal (default), json, jsonplus, or terse
        """

        self.fio_job = fio_job
        self.fio_pre_job = fio_pre_job
        self.fio_pre_success = fio_pre_success if fio_pre_success else success
        self.output_format = output_format
        self.precon_failed = False
        self.json_data = None

        super().__init__(fio_path, success, testnum, artifact_root)

    def setup(self, parameters):
        """Setup instance variables for fio job test."""

        self.filenames['fio_output'] = f"{os.path.basename(self.fio_job)}.output"
        fio_args = [
            "--max-jobs=16",
            f"--output-format={self.output_format}",
            f"--output={self.filenames['fio_output']}",
            self.fio_job,
            ]
        if parameters:
            fio_args += parameters

        super().setup(fio_args)

        # Update the filenames from the default
        self.filenames['cmd'] = os.path.join(self.paths['test_dir'],
                                             f"{os.path.basename(self.fio_job)}.command")
        self.filenames['stdout'] = os.path.join(self.paths['test_dir'],
                                                f"{os.path.basename(self.fio_job)}.stdout")
        self.filenames['stderr'] = os.path.join(self.paths['test_dir'],
                                                f"{os.path.basename(self.fio_job)}.stderr")
        self.filenames['exitcode'] = os.path.join(self.paths['test_dir'],
                                                  f"{os.path.basename(self.fio_job)}.exitcode")

    def run_pre_job(self):
        """Run fio job precondition step."""

        precon = FioJobFileTest(self.paths['exe'], self.fio_pre_job,
                            self.fio_pre_success,
                            self.testnum,
                            self.paths['artifacts'],
                            output_format=self.output_format)
        precon.setup(None)
        precon.run()
        precon.check_result()
        self.precon_failed = not precon.passed
        self.failure_reason = precon.failure_reason

    def run(self):
        """Run fio job test."""

        if self.fio_pre_job:
            self.run_pre_job()

        if not self.precon_failed:
            super().run()
        else:
            logging.debug("Test %d: precondition step failed", self.testnum)

    def get_file_fail(self, filename):
        """Safely read a file and fail the test upon error."""
        file_data = None

        try:
            with open(filename, "r", encoding=locale.getpreferredencoding()) as output_file:
                file_data = output_file.read()
        except OSError:
            self.failure_reason += f" unable to read file {filename}"
            self.passed = False

        return file_data

    def check_result(self):
        """Check fio job results."""

        if self.precon_failed:
            self.passed = False
            self.failure_reason = f"{self.failure_reason} precondition step failed,"
            return

        super().check_result()

        if not self.passed:
            return

        if 'json' not in self.output_format:
            return

        file_data = self.get_file_fail(os.path.join(self.paths['test_dir'],
                                                    self.filenames['fio_output']))
        if not file_data:
            return

        #
        # Sometimes fio informational messages are included at the top of the
        # JSON output, especially under Windows. Try to decode output as JSON
        # data, skipping everything until the first {
        #
        lines = file_data.splitlines()
        file_data = '\n'.join(lines[lines.index("{"):])
        try:
            self.json_data = json.loads(file_data)
        except json.JSONDecodeError:
            self.failure_reason = f"{self.failure_reason} unable to decode JSON data,"
            self.passed = False


class FioJobCmdTest(FioExeTest):
    """This runs a fio job with options specified on the command line."""

    def __init__(self, fio_path, success, testnum, artifact_root, fio_opts, basename=None):

        self.basename = basename if basename else os.path.basename(fio_path)
        self.fio_opts = fio_opts
        self.json_data = None
        self.iops_log_lines = None

        super().__init__(fio_path, success, testnum, artifact_root)

        filename_stub = os.path.join(self.paths['test_dir'], f"{self.basename}{self.testnum:03d}")
        self.filenames['cmd'] = f"{filename_stub}.command"
        self.filenames['stdout'] = f"{filename_stub}.stdout"
        self.filenames['stderr'] = f"{filename_stub}.stderr"
        self.filenames['output'] = os.path.abspath(f"{filename_stub}.output")
        self.filenames['exitcode'] = f"{filename_stub}.exitcode"
        self.filenames['iopslog'] = os.path.abspath(f"{filename_stub}")

    def run(self):
        super().run()

        if 'output-format' in self.fio_opts and 'json' in \
                self.fio_opts['output-format']:
            if not self.get_json():
                print('Unable to decode JSON data')
                self.passed = False

        if any('--write_iops_log=' in param for param in self.parameters):
            self.get_iops_log()

    def get_iops_log(self):
        """Read IOPS log from the first job."""

        log_filename = self.filenames['iopslog'] + "_iops.1.log"
        with open(log_filename, 'r', encoding=locale.getpreferredencoding()) as iops_file:
            self.iops_log_lines = iops_file.read()

    def get_json(self):
        """Convert fio JSON output into a python JSON object"""

        filename = self.filenames['output']
        with open(filename, 'r', encoding=locale.getpreferredencoding()) as file:
            file_data = file.read()

        #
        # Sometimes fio informational messages are included at the top of the
        # JSON output, especially under Windows. Try to decode output as JSON
        # data, lopping off up to the first four lines
        #
        lines = file_data.splitlines()
        for i in range(5):
            file_data = '\n'.join(lines[i:])
            try:
                self.json_data = json.loads(file_data)
            except json.JSONDecodeError:
                continue
            else:
                return True

        return False

    @staticmethod
    def check_empty(job):
        """
        Make sure JSON data is empty.

        Some data structures should be empty. This function makes sure that they are.

        job         JSON object that we need to check for emptiness
        """

        return job['total_ios'] == 0 and \
                job['slat_ns']['N'] == 0 and \
                job['clat_ns']['N'] == 0 and \
                job['lat_ns']['N'] == 0

    def check_all_ddirs(self, ddir_nonzero, job):
        """
        Iterate over the data directions and check whether each is
        appropriately empty or not.
        """

        retval = True
        ddirlist = ['read', 'write', 'trim']

        for ddir in ddirlist:
            if ddir in ddir_nonzero:
                if self.check_empty(job[ddir]):
                    print(f"Unexpected zero {ddir} data found in output")
                    retval = False
            else:
                if not self.check_empty(job[ddir]):
                    print(f"Unexpected {ddir} data found in output")
                    retval = False

        return retval


def run_fio_tests(test_list, test_env, args):
    """
    Run tests as specified in test_list.
    """

    passed = 0
    failed = 0
    skipped = 0

    for config in test_list:
        if (args.skip and config['test_id'] in args.skip) or \
           (args.run_only and config['test_id'] not in args.run_only) or \
           ('force_skip' in config and config['force_skip']):
            skipped = skipped + 1
            print(f"Test {config['test_id']} SKIPPED (User request or override)")
            continue

        if issubclass(config['test_class'], FioJobFileTest):
            if config['pre_job']:
                fio_pre_job = os.path.join(test_env['fio_root'], 't', 'jobs',
                                           config['pre_job'])
            else:
                fio_pre_job = None
            if config['pre_success']:
                fio_pre_success = config['pre_success']
            else:
                fio_pre_success = None
            if 'output_format' in config:
                output_format = config['output_format']
            else:
                output_format = 'normal'
            test = config['test_class'](
                test_env['fio_path'],
                os.path.join(test_env['fio_root'], 't', 'jobs', config['job']),
                config['success'],
                config['test_id'],
                test_env['artifact_root'],
                fio_pre_job=fio_pre_job,
                fio_pre_success=fio_pre_success,
                output_format=output_format)
            desc = config['job']
            parameters = config['parameters'] if 'parameters' in config else None
        elif issubclass(config['test_class'], FioJobCmdTest):
            if not 'success' in config:
                config['success'] = SUCCESS_DEFAULT
            test = config['test_class'](test_env['fio_path'],
                                        config['success'],
                                        config['test_id'],
                                        test_env['artifact_root'],
                                        config['fio_opts'],
                                        test_env['basename'])
            desc = config['test_id']
            parameters = config
        elif issubclass(config['test_class'], FioExeTest):
            exe_path = os.path.join(test_env['fio_root'], config['exe'])
            parameters = []
            if config['parameters']:
                parameters = [p.format(fio_path=test_env['fio_path'], nvmecdev=args.nvmecdev)
                              for p in config['parameters']]
            if Path(exe_path).suffix == '.py' and platform.system() == "Windows":
                parameters.insert(0, exe_path)
                exe_path = "python.exe"
            if config['test_id'] in test_env['pass_through']:
                parameters += test_env['pass_through'][config['test_id']].split()
            test = config['test_class'](
                    exe_path,
                    config['success'],
                    config['test_id'],
                    test_env['artifact_root'])
            desc = config['exe']
        else:
            print(f"Test {config['test_id']} FAILED: unable to process test config")
            failed = failed + 1
            continue

        if 'requirements' in config and not args.skip_req:
            reqs_met = True
            for req in config['requirements']:
                reqs_met, reason = req()
                logging.debug("Test %d: Requirement '%s' met? %s", config['test_id'], reason,
                              reqs_met)
                if not reqs_met:
                    break
            if not reqs_met:
                print(f"Test {config['test_id']} SKIPPED ({reason}) {desc}")
                skipped = skipped + 1
                continue

        try:
            test.setup(parameters)
            test.run()
            test.check_result()
        except KeyboardInterrupt:
            break
        except Exception as e:
            test.passed = False
            test.failure_reason += str(e)
            logging.debug("Test %d exception:\n%s\n", config['test_id'], traceback.format_exc())
        if test.passed:
            result = "PASSED"
            passed = passed + 1
        else:
            result = f"FAILED: {test.failure_reason}"
            failed = failed + 1
            contents, _ = get_file(test.filenames['stderr'])
            logging.debug("Test %d: stderr:\n%s", config['test_id'], contents)
            contents, _ = get_file(test.filenames['stdout'])
            logging.debug("Test %d: stdout:\n%s", config['test_id'], contents)
        print(f"Test {config['test_id']} {result} {desc}")

    print(f"{passed} test(s) passed, {failed} failed, {skipped} skipped")

    return passed, failed, skipped



#!/usr/bin/env python3
#
# Copyright 2024 Samsung Electronics Co., Ltd All Rights Reserved
#
# For conditions of distribution and use, see the accompanying COPYING file.
#
"""
# nvmept_trim.py
#
# Test fio's io_uring_cmd ioengine with NVMe pass-through dataset management
# commands that trim multiple ranges.
#
# USAGE
# see python3 nvmept_trim.py --help
#
# EXAMPLES
# python3 t/nvmept_trim.py --dut /dev/ng0n1
# python3 t/nvmept_trim.py --dut /dev/ng1n1 -f ./fio
#
# REQUIREMENTS
# Python 3.6
#
"""
import os
import sys
import time
import logging
import argparse
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests
from fiotestcommon import SUCCESS_NONZERO


class TrimTest(FioJobCmdTest):
    """
    NVMe pass-through test class. Check to make sure output for selected data
    direction(s) is non-zero and that zero data appears for other directions.
    """

    def setup(self, parameters):
        """Setup a test."""

        fio_args = [
            "--name=nvmept-trim",
            "--ioengine=io_uring_cmd",
            "--cmd_type=nvme",
            f"--filename={self.fio_opts['filename']}",
            f"--rw={self.fio_opts['rw']}",
            f"--output={self.filenames['output']}",
            f"--output-format={self.fio_opts['output-format']}",
        ]
        for opt in ['fixedbufs', 'nonvectored', 'force_async', 'registerfiles',
                    'sqthread_poll', 'sqthread_poll_cpu', 'hipri', 'nowait',
                    'time_based', 'runtime', 'verify', 'io_size', 'num_range',
                    'iodepth', 'iodepth_batch', 'iodepth_batch_complete',
                    'size', 'rate', 'bs', 'bssplit', 'bsrange', 'randrepeat',
                    'buffer_pattern', 'verify_pattern', 'verify', 'offset']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)


    def check_result(self):

        super().check_result()

        if 'rw' not in self.fio_opts or \
                not self.passed or \
                'json' not in self.fio_opts['output-format']:
            return

        job = self.json_data['jobs'][0]

        if self.fio_opts['rw'] in ['read', 'randread']:
            self.passed = self.check_all_ddirs(['read'], job)
        elif self.fio_opts['rw'] in ['write', 'randwrite']:
            if 'verify' not in self.fio_opts:
                self.passed = self.check_all_ddirs(['write'], job)
            else:
                self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif self.fio_opts['rw'] in ['trim', 'randtrim']:
            self.passed = self.check_all_ddirs(['trim'], job)
        elif self.fio_opts['rw'] in ['readwrite', 'randrw']:
            self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif self.fio_opts['rw'] in ['trimwrite', 'randtrimwrite']:
            self.passed = self.check_all_ddirs(['trim', 'write'], job)
        else:
            logging.error("Unhandled rw value %s", self.fio_opts['rw'])
            self.passed = False

        if 'iodepth' in self.fio_opts:
            # We will need to figure something out if any test uses an iodepth
            # different from 8
            if job['iodepth_level']['8'] < 95:
                logging.error("Did not achieve requested iodepth")
                self.passed = False
            else:
                logging.debug("iodepth 8 target met %s", job['iodepth_level']['8'])


class RangeTrimTest(TrimTest):
    """
    Multi-range trim test class.
    """

    def get_bs(self):
        """Calculate block size and determine whether bs will be an average or exact."""

        if 'bs' in self.fio_opts:
            exact_size = True
            bs = self.fio_opts['bs']
        elif 'bssplit' in self.fio_opts:
            exact_size = False
            bs = 0
            total = 0
            for split in self.fio_opts['bssplit'].split(':'):
                [blocksize, share] = split.split('/')
                total += int(share)
                bs += int(blocksize) * int(share) / 100
            if total != 100:
                logging.error("bssplit '%s' total percentage is not 100", self.fio_opts['bssplit'])
                self.passed = False
            else:
                logging.debug("bssplit: average block size is %d", int(bs))
            # The only check we do here for bssplit is to calculate an average
            # blocksize and see if the IOPS and bw are consistent
        elif 'bsrange' in self.fio_opts:
            exact_size = False
            [minbs, maxbs] = self.fio_opts['bsrange'].split('-')
            minbs = int(minbs)
            maxbs = int(maxbs)
            bs = int((minbs + maxbs) / 2)
            logging.debug("bsrange: average block size is %d", int(bs))
            # The only check we do here for bsrange is to calculate an average
            # blocksize and see if the IOPS and bw are consistent
        else:
            exact_size = True
            bs = 4096

        return bs, exact_size


    def check_result(self):
        """
        Make sure that the number of IO requests is consistent with the
        blocksize and num_range values. In other words, if the blocksize is
        4KiB and num_range is 2, we should have 128 IO requests to trim 1MiB.
        """
        # TODO Enable debug output to check the actual offsets

        super().check_result()

        if not self.passed or 'json' not in self.fio_opts['output-format']:
            return

        job = self.json_data['jobs'][0]['trim']
        bs, exact_size = self.get_bs()

        # make sure bw and IOPS are consistent
        bw = job['bw_bytes']
        iops = job['iops']
        runtime = job['runtime']

        calculated = int(bw*runtime/1000)
        expected = job['io_bytes']
        if abs(calculated - expected) / expected > 0.05:
            logging.error("Total bytes %d from bw does not match reported total bytes %d",
                          calculated, expected)
            self.passed = False
        else:
            logging.debug("Total bytes %d from bw matches reported total bytes %d", calculated,
                          expected)

        calculated = int(iops*runtime/1000*bs*self.fio_opts['num_range'])
        if abs(calculated - expected) / expected > 0.05:
            logging.error("Total bytes %d from IOPS does not match reported total bytes %d",
                          calculated, expected)
            self.passed = False
        else:
            logging.debug("Total bytes %d from IOPS matches reported total bytes %d", calculated,
                          expected)

        if 'size' in self.fio_opts:
            io_count = self.fio_opts['size'] / self.fio_opts['num_range'] / bs
            if exact_size:
                delta = 0.1
            else:
                delta = 0.05*job['total_ios']

            if abs(job['total_ios'] - io_count) > delta:
                logging.error("Expected numbers of IOs %d does not match actual value %d",
                              io_count, job['total_ios'])
                self.passed = False
            else:
                logging.debug("Expected numbers of IOs %d matches actual value %d", io_count,
                              job['total_ios'])

        if 'rate' in self.fio_opts:
            if abs(bw - self.fio_opts['rate']) / self.fio_opts['rate'] > 0.05:
                logging.error("Actual rate %f does not match expected rate %f", bw,
                              self.fio_opts['rate'])
                self.passed = False
            else:
                logging.debug("Actual rate %f matches expeected rate %f", bw, self.fio_opts['rate'])



TEST_LIST = [
    # The group of tests below checks existing use cases to make sure there are
    # no regressions.
    {
        "test_id": 1,
        "fio_opts": {
            "rw": 'trim',
            "time_based": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 2,
        "fio_opts": {
            "rw": 'randtrim',
            "time_based": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 3,
        "fio_opts": {
            "rw": 'trim',
            "time_based": 1,
            "runtime": 3,
            "iodepth": 8,
            "iodepth_batch": 4,
            "iodepth_batch_complete": 4,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 4,
        "fio_opts": {
            "rw": 'randtrim',
            "time_based": 1,
            "runtime": 3,
            "iodepth": 8,
            "iodepth_batch": 4,
            "iodepth_batch_complete": 4,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 5,
        "fio_opts": {
            "rw": 'trimwrite',
            "time_based": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 6,
        "fio_opts": {
            "rw": 'randtrimwrite',
            "time_based": 1,
            "runtime": 3,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 7,
        "fio_opts": {
            "rw": 'randtrim',
            "time_based": 1,
            "runtime": 3,
            "fixedbufs": 0,
            "nonvectored": 1,
            "force_async": 1,
            "registerfiles": 1,
            "sqthread_poll": 1,
            "fixedbuffs": 1,
            "output-format": "json",
            },
        "test_class": TrimTest,
    },
    # The group of tests below try out the new functionality
    {
        "test_id": 100,
        "fio_opts": {
            "rw": 'trim',
            "num_range": 2,
            "size": 16*1024*1024,
            "output-format": "json",
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 101,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 2,
            "size": 16*1024*1024,
            "output-format": "json",
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 102,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 256,
            "size": 64*1024*1024,
            "output-format": "json",
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 103,
        "fio_opts": {
            "rw": 'trim',
            "num_range": 2,
            "bs": 16*1024,
            "size": 32*1024*1024,
            "output-format": "json",
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 104,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 2,
            "bs": 16*1024,
            "size": 32*1024*1024,
            "output-format": "json",
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 105,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 2,
            "bssplit": "4096/50:16384/50",
            "size": 80*1024*1024,
            "output-format": "json",
            "randrepeat": 0,
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 106,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 4,
            "bssplit": "4096/25:8192/25:12288/25:16384/25",
            "size": 80*1024*1024,
            "output-format": "json",
            "randrepeat": 0,
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 107,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 4,
            "bssplit": "4096/20:8192/20:12288/20:16384/20:20480/20",
            "size": 72*1024*1024,
            "output-format": "json",
            "randrepeat": 0,
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 108,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 2,
            "bsrange": "4096-16384",
            "size": 80*1024*1024,
            "output-format": "json",
            "randrepeat": 0,
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 109,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 4,
            "bsrange": "4096-20480",
            "size": 72*1024*1024,
            "output-format": "json",
            "randrepeat": 0,
            },
        "test_class": RangeTrimTest,
    },
    {
        "test_id": 110,
        "fio_opts": {
            "rw": 'randtrim',
            "time_based": 1,
            "runtime": 10,
            "rate": 1024*1024,
            "num_range": 2,
            "output-format": "json",
            },
        "test_class": RangeTrimTest,
    },
    # All of the tests below should fail
    # TODO check the error messages resulting from the jobs below
    {
        "test_id": 200,
        "fio_opts": {
            "rw": 'randtrimwrite',
            "time_based": 1,
            "runtime": 10,
            "rate": 1024*1024,
            "num_range": 2,
            "output-format": "normal",
            },
        "test_class": RangeTrimTest,
        "success": SUCCESS_NONZERO,
    },
    {
        "test_id": 201,
        "fio_opts": {
            "rw": 'trimwrite',
            "time_based": 1,
            "runtime": 10,
            "rate": 1024*1024,
            "num_range": 2,
            "output-format": "normal",
            },
        "test_class": RangeTrimTest,
        "success": SUCCESS_NONZERO,
    },
    {
        "test_id": 202,
        "fio_opts": {
            "rw": 'trim',
            "time_based": 1,
            "runtime": 10,
            "num_range": 257,
            "output-format": "normal",
            },
        "test_class": RangeTrimTest,
        "success": SUCCESS_NONZERO,
    },
    # The sequence of jobs below constitute a single test with multiple steps
    # - write a data pattern
    # - verify the data pattern
    # - trim the first half of the LBA space
    # - verify that the trim'd LBA space no longer returns the original data pattern
    # - verify that the remaining LBA space has the expected pattern
    {
        "test_id": 300,
        "fio_opts": {
            "rw": 'write',
            "output-format": 'json',
            "buffer_pattern": 0x0f,
            "size": 256*1024*1024,
            "bs": 256*1024,
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 301,
        "fio_opts": {
            "rw": 'read',
            "output-format": 'json',
            "verify_pattern": 0x0f,
            "verify": "pattern",
            "size": 256*1024*1024,
            "bs": 256*1024,
            },
        "test_class": TrimTest,
    },
    {
        "test_id": 302,
        "fio_opts": {
            "rw": 'randtrim',
            "num_range": 8,
            "output-format": 'json',
            "size": 128*1024*1024,
            "bs": 256*1024,
            },
        "test_class": TrimTest,
    },
    # The identify namespace data structure has a DLFEAT field which specifies
    # what happens when reading data from deallocated blocks. There are three
    # options:
    # - read behavior not reported
    # - deallocated logical block returns all bytes 0x0
    # - deallocated logical block returns all bytes 0xff
    # The test below merely checks that the original data pattern is not returned.
    # Source: Figure 97 from
    # https://nvmexpress.org/wp-content/uploads/NVM-Express-NVM-Command-Set-Specification-1.0c-2022.10.03-Ratified.pdf
    {
        "test_id": 303,
        "fio_opts": {
            "rw": 'read',
            "output-format": 'json',
            "verify_pattern": 0x0f,
            "verify": "pattern",
            "size": 128*1024*1024,
            "bs": 256*1024,
            },
        "test_class": TrimTest,
        "success": SUCCESS_NONZERO,
    },
    {
        "test_id": 304,
        "fio_opts": {
            "rw": 'read',
            "output-format": 'json',
            "verify_pattern": 0x0f,
            "verify": "pattern",
            "offset": 128*1024*1024,
            "size": 128*1024*1024,
            "bs": 256*1024,
            },
        "test_class": TrimTest,
    },
]

def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', help='Enable debug messages', action='store_true')
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('--dut', help='target NVMe character device to test '
                        '(e.g., /dev/ng0n1). WARNING: THIS IS A DESTRUCTIVE TEST', required=True)
    args = parser.parse_args()

    return args


def main():
    """Run tests using fio's io_uring_cmd ioengine to send NVMe pass through commands."""

    args = parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    artifact_root = args.artifact_root if args.artifact_root else \
        f"nvmept-trim-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    for test in TEST_LIST:
        test['fio_opts']['filename'] = args.dut

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'nvmept-trim',
              }

    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (c) 2020 Western Digital Corporation or its affiliates.
#
"""
jsonplus2csv-test.py

Do one basic test of tools/fio_jsonplus2csv

USAGE
python jsonplus2csv-test.py [-f fio-executable] [-s script-location]

EXAMPLES
python t/jsonplus2csv-test.py
python t/jsonplus2csv-test.py -f ./fio -s tools

REQUIREMENTS
Python 3.5+
"""

import os
import sys
import platform
import argparse
import subprocess


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--fio',
                        help='path to fio executable (e.g., ./fio)')
    parser.add_argument('-s', '--script',
                        help='directory containing fio_jsonplus2csv script')
    return parser.parse_args()


def run_fio(fio):
    """Run fio to generate json+ data.

    Parameters:
        fio     path to fio executable.
    """

# We need an async ioengine to get submission latencies
    if platform.system() == 'Linux':
        aio = 'libaio'
    elif platform.system() == 'Windows':
        aio = 'windowsaio'
    else:
        aio = 'posixaio'

    fio_args = [
        "--max-jobs=4",
        "--output=fio-output.json",
        "--output-format=json+",
        "--filename=fio_jsonplus_clat2csv.test",
        "--ioengine=" + aio,
        "--time_based",
        "--runtime=3s",
        "--size=1M",
        "--slat_percentiles=1",
        "--clat_percentiles=1",
        "--lat_percentiles=1",
        "--thread=1",
        "--name=test1",
        "--rw=randrw",
        "--name=test2",
        "--rw=read",
        "--name=test3",
        "--rw=write",
        ]

    output = subprocess.run([fio] + fio_args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    return output


def check_output(fio_output, script_path):
    """Run t/fio_jsonplus_clat2csv and validate the generated CSV files
    against the original json+ fio output.

    Parameters:
        fio_output      subprocess.run object describing fio run.
        script_path     path to fio_jsonplus_clat2csv script.
    """

    if fio_output.returncode != 0:
        print("ERROR: fio run failed")
        return False

    if platform.system() == 'Windows':
        script = ['python.exe', script_path]
    else:
        script = [script_path]

    script_args = ["fio-output.json", "fio-output.csv"]
    script_args_validate = script_args + ["--validate"]

    script_output = subprocess.run(script + script_args)
    if script_output.returncode != 0:
        return False

    script_output = subprocess.run(script + script_args_validate)
    if script_output.returncode != 0:
        return False

    return True


def main():
    """Entry point for this script."""

    args = parse_args()

    index = 1
    passed = 0
    failed = 0

    if args.fio:
        fio_path = args.fio
    else:
        fio_path = os.path.join(os.path.dirname(__file__), '../fio')
        if not os.path.exists(fio_path):
            fio_path = 'fio'
    print("fio path is", fio_path)

    if args.script:
        script_path = args.script
    else:
        script_path = os.path.join(os.path.dirname(__file__), '../tools/fio_jsonplus_clat2csv')
        if not os.path.exists(script_path):
            script_path = 'fio_jsonplus_clat2csv'
    print("script path is", script_path)

    fio_output = run_fio(fio_path)
    status = check_output(fio_output, script_path)
    print("Test {0} {1}".format(index, ("PASSED" if status else "FAILED")))
    if status:
        passed = passed + 1
    else:
        failed = failed + 1
    index = index + 1

    print("{0} tests passed, {1} failed".format(passed, failed))

    sys.exit(failed)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
fiotestcommon.py

This contains constant definitions, helpers, and a Requirements class that can
be used to help with running fio tests.
"""

import os
import locale
import logging
import platform
import subprocess
import multiprocessing


SUCCESS_DEFAULT = {
    'zero_return': True,
    'stderr_empty': True,
    'timeout': 600,
    }
SUCCESS_NONZERO = {
    'zero_return': False,
    'stderr_empty': False,
    'timeout': 600,
    }
SUCCESS_STDERR = {
    'zero_return': True,
    'stderr_empty': False,
    'timeout': 600,
    }


def get_file(filename):
    """Safely read a file."""
    file_data = ''
    success = True

    try:
        with open(filename, "r", encoding=locale.getpreferredencoding()) as output_file:
            file_data = output_file.read()
    except OSError:
        success = False

    return file_data, success


class Requirements():
    """Requirements consists of multiple run environment characteristics.
    These are to determine if a particular test can be run"""

    _linux = False
    _libaio = False
    _io_uring = False
    _zbd = False
    _root = False
    _zoned_nullb = False
    _not_macos = False
    _not_windows = False
    _unittests = False
    _cpucount4 = False
    _nvmecdev = False

    def __init__(self, fio_root, args):
        Requirements._not_macos = platform.system() != "Darwin"
        Requirements._not_windows = platform.system() != "Windows"
        Requirements._linux = platform.system() == "Linux"

        if Requirements._linux:
            config_file = os.path.join(fio_root, "config-host.h")
            contents, success = get_file(config_file)
            if not success:
                print(f"Unable to open {config_file} to check requirements")
                Requirements._zbd = True
            else:
                Requirements._zbd = "CONFIG_HAS_BLKZONED" in contents
                Requirements._libaio = "CONFIG_LIBAIO" in contents

            contents, success = get_file("/proc/kallsyms")
            if not success:
                print("Unable to open '/proc/kallsyms' to probe for io_uring support")
            else:
                Requirements._io_uring = "io_uring_setup" in contents

            Requirements._root = os.geteuid() == 0
            if Requirements._zbd and Requirements._root:
                try:
                    subprocess.run(["modprobe", "null_blk"],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
                    if os.path.exists("/sys/module/null_blk/parameters/zoned"):
                        Requirements._zoned_nullb = True
                except Exception:
                    pass

        if platform.system() == "Windows":
            utest_exe = "unittest.exe"
        else:
            utest_exe = "unittest"
        unittest_path = os.path.join(fio_root, "unittests", utest_exe)
        Requirements._unittests = os.path.exists(unittest_path)

        Requirements._cpucount4 = multiprocessing.cpu_count() >= 4
        Requirements._nvmecdev = args.nvmecdev

        req_list = [
                Requirements.linux,
                Requirements.libaio,
                Requirements.io_uring,
                Requirements.zbd,
                Requirements.root,
                Requirements.zoned_nullb,
                Requirements.not_macos,
                Requirements.not_windows,
                Requirements.unittests,
                Requirements.cpucount4,
                Requirements.nvmecdev,
                    ]
        for req in req_list:
            value, desc = req()
            logging.debug("Requirements: Requirement '%s' met? %s", desc, value)

    @classmethod
    def linux(cls):
        """Are we running on Linux?"""
        return Requirements._linux, "Linux required"

    @classmethod
    def libaio(cls):
        """Is libaio available?"""
        return Requirements._libaio, "libaio required"

    @classmethod
    def io_uring(cls):
        """Is io_uring available?"""
        return Requirements._io_uring, "io_uring required"

    @classmethod
    def zbd(cls):
        """Is ZBD support available?"""
        return Requirements._zbd, "Zoned block device support required"

    @classmethod
    def root(cls):
        """Are we running as root?"""
        return Requirements._root, "root required"

    @classmethod
    def zoned_nullb(cls):
        """Are zoned null block devices available?"""
        return Requirements._zoned_nullb, "Zoned null block device support required"

    @classmethod
    def not_macos(cls):
        """Are we running on a platform other than macOS?"""
        return Requirements._not_macos, "platform other than macOS required"

    @classmethod
    def not_windows(cls):
        """Are we running on a platform other than Windws?"""
        return Requirements._not_windows, "platform other than Windows required"

    @classmethod
    def unittests(cls):
        """Were unittests built?"""
        return Requirements._unittests, "Unittests support required"

    @classmethod
    def cpucount4(cls):
        """Do we have at least 4 CPUs?"""
        return Requirements._cpucount4, "4+ CPUs required"

    @classmethod
    def nvmecdev(cls):
        """Do we have an NVMe character device to test?"""
        return Requirements._nvmecdev, "NVMe character device test target required"



#!/usr/bin/env python3
""" Cut the number bins in half in fio histogram output. Example usage:

        $ half-bins.py -c 2 output_clat_hist.1.log > smaller_clat_hist.1.log

    Which merges e.g. bins [0 .. 3], [4 .. 7], ..., [1212 .. 1215] resulting in
    304 = 1216 / (2**2) merged bins per histogram sample.

    @author Karl Cronburg <karl.cronburg@gmail.com>
"""
import sys

def main(ctx):
    stride = 1 << ctx.coarseness
    with open(ctx.FILENAME, 'r') as fp:
        for line in fp.readlines():
            vals = line.split(', ')
            sys.stdout.write("%s, %s, %s, " % tuple(vals[:3]))

            hist = list(map(int, vals[3:]))
            for i in range(0, len(hist) - stride, stride):
                sys.stdout.write("%d, " % sum(hist[i : i + stride],))
            sys.stdout.write("%d\n" % sum(hist[len(hist) - stride:]))

if __name__ == '__main__':
    import argparse
    p = argparse.ArgumentParser()
    arg = p.add_argument
    arg( 'FILENAME', help='clat_hist file for which we will reduce'
                         ' (by half or more) the number of bins.')
    arg('-c', '--coarseness',
       default=1,
       type=int,
       help='number of times to reduce number of bins by half, '
            'e.g. coarseness of 4 merges each 2^4 = 16 consecutive '
            'bins.')
    main(p.parse_args())



#!/usr/bin/env python3
#
# Copyright 2024 Samsung Electronics Co., Ltd All Rights Reserved
#
# For conditions of distribution and use, see the accompanying COPYING file.
#
"""
# nvmept_fdp.py
#
# Test fio's io_uring_cmd ioengine with NVMe pass-through FDP write commands.
#
# USAGE
# see python3 nvmept_fdp.py --help
#
# EXAMPLES
# python3 t/nvmept_fdp.py --dut /dev/ng0n1
# python3 t/nvmept_fdp.py --dut /dev/ng1n1 -f ./fio
#
# REQUIREMENTS
# Python 3.6
# Device formatted with LBA data size 4096 bytes
# Device with at least five placement IDs
#
# WARNING
# This is a destructive test
"""
import os
import sys
import json
import time
import locale
import logging
import argparse
import subprocess
from pathlib import Path
from fiotestlib import FioJobCmdTest, run_fio_tests
from fiotestcommon import SUCCESS_NONZERO

# This needs to match FIO_MAX_DP_IDS and DP_MAX_SCHEME_ENTRIES in
# dataplacement.h
FIO_MAX_DP_IDS = 128
DP_MAX_SCHEME_ENTRIES = 32

class FDPTest(FioJobCmdTest):
    """
    NVMe pass-through test class. Check to make sure output for selected data
    direction(s) is non-zero and that zero data appears for other directions.
    """

    def setup(self, parameters):
        """Setup a test."""

        fio_args = [
            "--name=nvmept-fdp",
            "--ioengine=io_uring_cmd",
            "--cmd_type=nvme",
            "--randrepeat=0",
            f"--filename={self.fio_opts['filename']}",
            f"--rw={self.fio_opts['rw']}",
            f"--output={self.filenames['output']}",
            f"--output-format={self.fio_opts['output-format']}",
        ]

        for opt in ['fixedbufs', 'nonvectored', 'force_async', 'registerfiles',
                    'sqthread_poll', 'sqthread_poll_cpu', 'hipri', 'nowait',
                    'time_based', 'runtime', 'verify', 'io_size', 'num_range',
                    'iodepth', 'iodepth_batch', 'iodepth_batch_complete',
                    'size', 'rate', 'bs', 'bssplit', 'bsrange', 'randrepeat',
                    'buffer_pattern', 'verify_pattern', 'offset', 'fdp',
                    'fdp_pli', 'fdp_pli_select', 'dataplacement', 'plid_select',
                    'plids', 'dp_scheme', 'number_ios', 'read_iolog']:
            if opt in self.fio_opts:
                option = f"--{opt}={self.fio_opts[opt]}"
                fio_args.append(option)

        super().setup(fio_args)


    def check_result(self):
        try:
            self._check_result()
        finally:
            if not update_all_ruhs(self.fio_opts['filename']):
                logging.error("Could not reset device")
            if not check_all_ruhs(self.fio_opts['filename']):
                logging.error("Reclaim units have inconsistent RUAMW values")


    def _check_result(self):

        super().check_result()

        if 'rw' not in self.fio_opts or \
                not self.passed or \
                'json' not in self.fio_opts['output-format']:
            return

        job = self.json_data['jobs'][0]
        rw_fio_opts = self.fio_opts['rw'].split(':')[0]

        if rw_fio_opts in ['read', 'randread']:
            self.passed = self.check_all_ddirs(['read'], job)
        elif rw_fio_opts in ['write', 'randwrite']:
            if 'verify' not in self.fio_opts:
                self.passed = self.check_all_ddirs(['write'], job)
            else:
                self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif rw_fio_opts in ['trim', 'randtrim']:
            self.passed = self.check_all_ddirs(['trim'], job)
        elif rw_fio_opts in ['readwrite', 'randrw']:
            self.passed = self.check_all_ddirs(['read', 'write'], job)
        elif rw_fio_opts in ['trimwrite', 'randtrimwrite']:
            self.passed = self.check_all_ddirs(['trim', 'write'], job)
        else:
            logging.error("Unhandled rw value %s", self.fio_opts['rw'])
            self.passed = False

        if 'iodepth' in self.fio_opts:
            # We will need to figure something out if any test uses an iodepth
            # different from 8
            if job['iodepth_level']['8'] < 95:
                logging.error("Did not achieve requested iodepth")
                self.passed = False
            else:
                logging.debug("iodepth 8 target met %s", job['iodepth_level']['8'])


class FDPMultiplePLIDTest(FDPTest):
    """
    Write to multiple placement IDs.
    """

    def setup(self, parameters):
        mapping = {
                    'nruhsd': FIO_FDP_NUMBER_PLIDS,
                    'max_ruamw': FIO_FDP_MAX_RUAMW,
                    'maxplid': FIO_FDP_NUMBER_PLIDS-1,
                    # parameters for 400, 401 tests
                    'hole_size': 64*1024,
                    'nios_for_scheme': min(FIO_FDP_NUMBER_PLIDS//2, DP_MAX_SCHEME_ENTRIES),
                }
        if 'number_ios' in self.fio_opts and isinstance(self.fio_opts['number_ios'], str):
            self.fio_opts['number_ios'] = eval(self.fio_opts['number_ios'].format(**mapping))
        if 'bs' in self.fio_opts and isinstance(self.fio_opts['bs'], str):
            self.fio_opts['bs'] = eval(self.fio_opts['bs'].format(**mapping))
        if 'rw' in self.fio_opts and isinstance(self.fio_opts['rw'], str):
            self.fio_opts['rw'] = self.fio_opts['rw'].format(**mapping)
        if 'plids' in self.fio_opts and isinstance(self.fio_opts['plids'], str):
            self.fio_opts['plids'] = self.fio_opts['plids'].format(**mapping)
        if 'fdp_pli' in self.fio_opts and isinstance(self.fio_opts['fdp_pli'], str):
            self.fio_opts['fdp_pli'] = self.fio_opts['fdp_pli'].format(**mapping)

        super().setup(parameters)
        
        if 'dp_scheme' in self.fio_opts:
            scheme_path = os.path.join(self.paths['test_dir'], self.fio_opts['dp_scheme'])
            with open(scheme_path, mode='w') as f:
                for i in range(mapping['nios_for_scheme']):
                    f.write(f'{mapping["hole_size"] * 2 * i}, {mapping["hole_size"] * 2 * (i+1)}, {i}\n')

        if 'read_iolog' in self.fio_opts:
            read_iolog_path = os.path.join(self.paths['test_dir'], self.fio_opts['read_iolog'])
            with open(read_iolog_path, mode='w') as f:
                f.write('fio version 2 iolog\n')
                f.write(f'{self.fio_opts["filename"]} add\n')
                f.write(f'{self.fio_opts["filename"]} open\n')

                for i in range(mapping['nios_for_scheme']):
                    f.write(f'{self.fio_opts["filename"]} write {mapping["hole_size"] * 2 * i} {mapping["hole_size"]}\n')

                f.write(f'{self.fio_opts["filename"]} close')
 
    def _check_result(self):
        if 'fdp_pli' in self.fio_opts:
            plid_list = self.fio_opts['fdp_pli'].split(',')
        elif 'plids' in self.fio_opts:
            plid_list = self.fio_opts['plids'].split(',')
        else:
            plid_list = [str(i) for i in range(FIO_FDP_NUMBER_PLIDS)]

        range_ids = []
        for plid in plid_list:
            if '-' in plid:
                [start, end] = plid.split('-')
                range_ids.extend(list(range(int(start), int(end)+1)))
            else:
                range_ids.append(int(plid))

        plid_list = sorted(range_ids)
        logging.debug("plid_list: %s", str(plid_list))

        fdp_status = get_fdp_status(self.fio_opts['filename'])

        select = "roundrobin"
        if 'fdp_pli_select' in self.fio_opts:
            select = self.fio_opts['fdp_pli_select']
        elif 'plid_select' in self.fio_opts:
            select = self.fio_opts['plid_select']

        if select == "roundrobin":
            self._check_robin(plid_list, fdp_status)
        elif select == "random":
            self._check_random(plid_list, fdp_status)
        elif select == "scheme":
            self._check_scheme(plid_list, fdp_status)
        else:
            logging.error("Unknown plid selection strategy %s", select)
            self.passed = False
        
        super()._check_result()

    def _check_robin(self, plid_list, fdp_status):
        """
        With round robin we can know exactly how many writes each PLID will
        receive.
        """
        ruamw = [FIO_FDP_MAX_RUAMW] * FIO_FDP_NUMBER_PLIDS

        number_ios = self.fio_opts['number_ios'] % (len(plid_list)*FIO_FDP_MAX_RUAMW)
        remainder = int(number_ios % len(plid_list))
        whole = int((number_ios - remainder) / len(plid_list))
        logging.debug("PLIDs in the list should show they have received %d writes; %d PLIDs will receive one extra",
                      whole, remainder)

        for plid in plid_list:
            ruamw[plid] -= whole
            if remainder:
                ruamw[plid] -= 1
                remainder -= 1
        logging.debug("Expected ruamw values: %s", str(ruamw))

        for idx, ruhs in enumerate(fdp_status['ruhss']):
            if idx >= FIO_FDP_NUMBER_PLIDS:
                break

            if ruhs['ruamw'] != ruamw[idx]:
                logging.error("RUAMW mismatch with idx %d, pid %d, expected %d, observed %d", idx,
                              ruhs['pid'], ruamw[idx], ruhs['ruamw'])
                self.passed = False
                break

            logging.debug("RUAMW match with idx %d, pid %d: ruamw=%d", idx, ruhs['pid'], ruamw[idx])

    def _check_random(self, plid_list, fdp_status):
        """
        With random selection, a set of PLIDs will receive all the write
        operations and the remainder will be untouched.
        """

        total_ruamw = 0
        for plid in plid_list:
            total_ruamw += fdp_status['ruhss'][plid]['ruamw']

        expected = len(plid_list) * FIO_FDP_MAX_RUAMW - self.fio_opts['number_ios']
        if total_ruamw != expected:
            logging.error("Expected total ruamw %d for plids %s, observed %d", expected,
                          str(plid_list), total_ruamw)
            self.passed = False
        else:
            logging.debug("Observed expected total ruamw %d for plids %s", expected, str(plid_list))

        for idx, ruhs in enumerate(fdp_status['ruhss']):
            if idx in plid_list:
                continue
            if ruhs['ruamw'] != FIO_FDP_MAX_RUAMW:
                logging.error("Unexpected ruamw %d for idx %d, pid %d, expected %d", ruhs['ruamw'],
                              idx, ruhs['pid'], FIO_FDP_MAX_RUAMW)
                self.passed = False
            else:
                logging.debug("Observed expected ruamw %d for idx %d, pid %d", ruhs['ruamw'], idx,
                              ruhs['pid'])

    def _check_scheme(self, plid_list, fdp_status):
        """
        With scheme selection, a set of PLIDs touched by the scheme
        """

        PLID_IDX_POS = 2
        plid_list_from_scheme = set()

        scheme_path = os.path.join(self.paths['test_dir'], self.fio_opts['dp_scheme'])

        with open(scheme_path) as f:
            lines = f.readlines()
            for line in lines:
                line_elem = line.strip().replace(' ', '').split(',')
                plid_list_from_scheme.add(int(line_elem[PLID_IDX_POS]))

        logging.debug(f'plid_list_from_scheme: {plid_list_from_scheme}')

        for idx, ruhs in enumerate(fdp_status['ruhss']):
            if ruhs['pid'] in plid_list_from_scheme:
                if ruhs['ruamw'] == FIO_FDP_MAX_RUAMW:
                    logging.error("pid %d should be touched by the scheme. But ruamw of it(%d) equals to %d",
                                    ruhs['pid'], ruhs['ruamw'], FIO_FDP_MAX_RUAMW)
                    self.passed = False
                else:
                    logging.debug("pid %d should be touched by the scheme. ruamw of it(%d) is under %d",
                                    ruhs['pid'], ruhs['ruamw'], FIO_FDP_MAX_RUAMW)
            else:
                if ruhs['ruamw'] == FIO_FDP_MAX_RUAMW:
                    logging.debug("pid %d should not be touched by the scheme. ruamw of it(%d) equals to %d",
                                    ruhs['pid'], ruhs['ruamw'], FIO_FDP_MAX_RUAMW)
                else:
                    logging.error("pid %d should not be touched by the scheme. But ruamw of it(%d) is under %d",
                                    ruhs['pid'], ruhs['ruamw'], FIO_FDP_MAX_RUAMW)
                    self.passed = False


class FDPSinglePLIDTest(FDPTest):
    """
    Write to a single placement ID only.
    """

    def _check_result(self):
        if 'plids' in self.fio_opts:
            plid = self.fio_opts['plids']
        elif 'fdp_pli' in self.fio_opts:
            plid = self.fio_opts['fdp_pli']
        else:
            plid = 0

        fdp_status = get_fdp_status(self.fio_opts['filename'])
        ruamw = fdp_status['ruhss'][plid]['ruamw']
        lba_count = self.fio_opts['number_ios']

        if FIO_FDP_MAX_RUAMW - lba_count != ruamw:
            logging.error("FDP accounting mismatch for plid %d; expected ruamw %d, observed %d",
                          plid, FIO_FDP_MAX_RUAMW - lba_count, ruamw)
            self.passed = False
        else:
            logging.debug("FDP accounting as expected for plid %d; ruamw = %d", plid, ruamw)

        super()._check_result()


class FDPReadTest(FDPTest):
    """
    Read workload test.
    """

    def _check_result(self):
        ruamw = check_all_ruhs(self.fio_opts['filename'])

        if ruamw != FIO_FDP_MAX_RUAMW:
            logging.error("Read workload affected FDP ruamw")
            self.passed = False
        else:
            logging.debug("Read workload did not disturb FDP ruamw")
            super()._check_result()


def get_fdp_status(dut):
    """
    Run the nvme-cli command to obtain FDP status and return result as a JSON
    object.
    """

    cmd = f"sudo nvme fdp status --output-format=json {dut}"
    cmd = cmd.split(' ')
    cmd_result = subprocess.run(cmd, capture_output=True, check=False,
                                encoding=locale.getpreferredencoding())

    if cmd_result.returncode != 0:
        logging.error("Error obtaining device %s FDP status: %s", dut, cmd_result.stderr)
        return False

    return json.loads(cmd_result.stdout)


def update_ruh(dut, plid):
    """
    Update reclaim unit handles with specified ID(s). This tells the device to
    point the RUH to a new (empty) reclaim unit.
    """

    ids = ','.join(plid) if isinstance(plid, list) else plid
    cmd = f"nvme fdp update --pids={ids} {dut}"
    cmd = cmd.split(' ')
    cmd_result = subprocess.run(cmd, capture_output=True, check=False,
                                encoding=locale.getpreferredencoding())

    if cmd_result.returncode != 0:
        logging.error("Error updating RUH %s ID(s) %s", dut, ids)
        return False

    return True


def update_all_ruhs(dut):
    """
    Update all reclaim unit handles on the device.
    """

    fdp_status = get_fdp_status(dut)
    for ruhs in fdp_status['ruhss']:
        if not update_ruh(dut, ruhs['pid']):
            return False

    return True


def check_all_ruhs(dut):
    """
    Check that all RUHs have the same value for reclaim unit available media
    writes (RUAMW).  Return the RUAMW value.
    """

    fdp_status = get_fdp_status(dut)
    ruh_status = fdp_status['ruhss']

    ruamw = ruh_status[0]['ruamw']
    for ruhs in ruh_status:
        if ruhs['ruamw'] != ruamw:
            logging.error("RUAMW mismatch: found %d, expected %d", ruhs['ruamw'], ruamw)
            return False

    return ruamw


TEST_LIST = [
    # Write one LBA to one PLID using both the old and new sets of options
    ## omit fdp_pli_select/plid_select
    {
        "test_id": 1,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 1,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 3,
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    {
        "test_id": 2,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 1,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 3,
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    ## fdp_pli_select/plid_select=roundrobin
    {
        "test_id": 3,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 1,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 3,
            "fdp_pli_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    {
        "test_id": 4,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 1,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 3,
            "plid_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    ## fdp_pli_select/plid_select=random
    {
        "test_id": 5,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 1,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 3,
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    {
        "test_id": 6,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 1,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 3,
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    # Write four LBAs to one PLID using both the old and new sets of options
    ## omit fdp_pli_select/plid_select
    {
        "test_id": 7,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 4,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 1,
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    {
        "test_id": 8,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 4,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 1,
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    ## fdp_pli_select/plid_select=roundrobin
    {
        "test_id": 9,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 4,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 1,
            "fdp_pli_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    {
        "test_id": 10,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 4,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 1,
            "plid_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    ## fdp_pli_select/plid_select=random
    {
        "test_id": 11,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "number_ios": 4,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 1,
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    {
        "test_id": 12,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 4,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 1,
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    # Just a regular write without FDP directive--should land on plid 0
    {
        "test_id": 13,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": 19,
            "verify": "crc32c",
            "output-format": "json",
            },
        "test_class": FDPSinglePLIDTest,
    },
    # Read workload
    {
        "test_id": 14,
        "fio_opts": {
            "rw": 'randread',
            "bs": 4096,
            "number_ios": 19,
            "output-format": "json",
            },
        "test_class": FDPReadTest,
    },
    # write to multiple PLIDs using round robin to select PLIDs
    ## write to all PLIDs using old and new sets of options
    {
        "test_id": 100,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "2*{nruhsd}+3",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 101,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "2*{nruhsd}+3",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plid_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    ## write to a subset of PLIDs using old and new sets of options
    {
        "test_id": 102,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{nruhsd}+1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "1,3",
            "fdp_pli_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 103,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{nruhsd}+1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": "1,3",
            "plid_select": "roundrobin",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    # write to multiple PLIDs using random selection of PLIDs
    ## write to all PLIDs using old and new sets of options
    {
        "test_id": 200,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 201,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    ## write to a subset of PLIDs using old and new sets of options
    {
        "test_id": 202,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "1,3,4",
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 203,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": "1,3,4",
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    ### use 3-4 to specify plids
    {
        "test_id": 204,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "1,3-4",
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 205,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": "1,3-4",
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    ### use 1-3 to specify plids
    {
        "test_id": 206,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "1-3",
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 207,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": "1-3",
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    ### use multiple ranges to specify plids
    {
        "test_id": 208,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "1-2,3-3",
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 209,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": "1-2,3-3",
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 210,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "0-{maxplid}",
            "fdp_pli_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 211,
        "fio_opts": {
            "rw": 'randwrite',
            "bs": 4096,
            "number_ios": "{max_ruamw}-1",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "fdp_pli": "0-{maxplid}",
            "plid_select": "random",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    # Specify invalid options fdp=1 and dataplacement=none
    {
        "test_id": 300,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 3,
            "output-format": "normal",
            "dataplacement": "none",
            },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    # Specify invalid options fdp=1 and dataplacement=streams
    {
        "test_id": 301,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 3,
            "output-format": "normal",
            "dataplacement": "streams",
            },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    # Specify invalid options related to dataplacement scheme
    ## using old and new sets of options
    {
        "test_id": 302,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": 3,
            "fdp_pli_select": "scheme",
            "output-format": "normal",
        },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    {
        "test_id": 303,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plids": 3,
            "plid_select": "scheme",
            "output-format": "normal",
        },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    ## Specify invalid ranges with start > end
    {
        "test_id": 304,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "plids": "3-1",
            "output-format": "normal",
            },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    {
        "test_id": 305,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "3-1",
            "output-format": "normal",
            },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    ## Specify too many plids
    {
        "test_id": 306,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "plids": "0-65535",
            "output-format": "normal",
            },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    {
        "test_id": 307,
        "fio_opts": {
            "rw": 'write',
            "bs": 4096,
            "io_size": 4096,
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli": "0-65535",
            "output-format": "normal",
            },
        "test_class": FDPTest,
        "success": SUCCESS_NONZERO,
    },
    # write to multiple PLIDs using scheme selection of PLIDs
    ## using old and new sets of options
    {
        "test_id": 400,
        "fio_opts": {
            "rw": "write:{hole_size}",
            "bs": "{hole_size}",
            "number_ios": "{nios_for_scheme}",
            "verify": "crc32c",
            "fdp": 1,
            "fdp_pli_select": "scheme",
            "dp_scheme": "lba.scheme",            
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    {
        "test_id": 401,
        "fio_opts": {
            "rw": "write:{hole_size}",
            "bs": "{hole_size}",
            "number_ios": "{nios_for_scheme}",
            "verify": "crc32c",
            "dataplacement": "fdp",
            "plid_select": "scheme",
            "dp_scheme": "lba.scheme",
            "output-format": "json",
            },
        "test_class": FDPMultiplePLIDTest,
    },
    # check whether dataplacement works while replaying iologs
    {
        "test_id": 402,
        "fio_opts": {
            "rw": "write:{hole_size}",
            "bs": "{hole_size}",
            "number_ios": "{nios_for_scheme}",
            "verify": "crc32c",
            "read_iolog": "iolog",
            "dataplacement": "fdp",
            "plid_select": "scheme",
            "dp_scheme": "lba.scheme",
            "output-format": "json",
        },
        "test_class": FDPMultiplePLIDTest,
    },
]

def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', help='Enable debug messages', action='store_true')
    parser.add_argument('-f', '--fio', help='path to file executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root', help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('--dut', help='target NVMe character device to test '
                        '(e.g., /dev/ng0n1). WARNING: THIS IS A DESTRUCTIVE TEST', required=True)
    args = parser.parse_args()

    return args


FIO_FDP_MAX_RUAMW = 0
FIO_FDP_NUMBER_PLIDS = 0

def main():
    """Run tests using fio's io_uring_cmd ioengine to send NVMe pass through commands."""
    global FIO_FDP_MAX_RUAMW
    global FIO_FDP_NUMBER_PLIDS

    args = parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    artifact_root = args.artifact_root if args.artifact_root else \
        f"nvmept-fdp-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if args.fio:
        fio_path = str(Path(args.fio).absolute())
    else:
        fio_path = 'fio'
    print(f"fio path is {fio_path}")

    for test in TEST_LIST:
        test['fio_opts']['filename'] = args.dut

    fdp_status = get_fdp_status(args.dut)
    FIO_FDP_NUMBER_PLIDS = min(fdp_status['nruhsd'], 128)
    update_all_ruhs(args.dut)
    FIO_FDP_MAX_RUAMW = check_all_ruhs(args.dut)
    if not FIO_FDP_MAX_RUAMW:
        sys.exit(-1)

    test_env = {
              'fio_path': fio_path,
              'fio_root': str(Path(__file__).absolute().parent.parent),
              'artifact_root': artifact_root,
              'basename': 'nvmept-fdp',
              }

    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only
#
# Copyright (c) 2019 Western Digital Corporation or its affiliates.
#
"""
# run-fio-tests.py
#
# Automate running of fio tests
#
# USAGE
# python3 run-fio-tests.py [-r fio-root] [-f fio-path] [-a artifact-root]
#                           [--skip # # #...] [--run-only # # #...]
#
#
# EXAMPLE
# # git clone git://git.kernel.dk/fio.git
# # cd fio
# # make -j
# # python3 t/run-fio-tests.py
#
#
# REQUIREMENTS
# - Python 3.5 (subprocess.run)
# - Linux (libaio ioengine, zbd tests, etc)
# - The artifact directory must be on a file system that accepts 512-byte IO
#   (t0002, t0003, t0004).
# - The artifact directory needs to be on an SSD. Otherwise tests that carry
#   out file-based IO will trigger a timeout (t0006).
# - 4 CPUs (t0009)
# - SciPy (steadystate_tests.py)
# - libzbc (zbd tests)
# - root privileges (zbd test)
# - kernel 4.19 or later for zoned null block devices (zbd tests)
# - CUnit support (unittests)
#
"""

#
# TODO  run multiple tests simultaneously
# TODO  Add sgunmap tests (requires SAS SSD)
#

import os
import sys
import time
import shutil
import logging
import argparse
import re
from pathlib import Path
from statsmodels.sandbox.stats.runs import runstest_1samp
from fiotestlib import FioExeTest, FioJobFileTest, run_fio_tests
from fiotestcommon import *


class FioJobFileTest_t0005(FioJobFileTest):
    """Test consists of fio test job t0005
    Confirm that read['io_kbytes'] == write['io_kbytes'] == 102400"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        if self.json_data['jobs'][0]['read']['io_kbytes'] != 102400:
            self.failure_reason = f"{self.failure_reason} bytes read mismatch,"
            self.passed = False
        if self.json_data['jobs'][0]['write']['io_kbytes'] != 102400:
            self.failure_reason = f"{self.failure_reason} bytes written mismatch,"
            self.passed = False


class FioJobFileTest_t0006(FioJobFileTest):
    """Test consists of fio test job t0006
    Confirm that read['io_kbytes'] ~ 2*write['io_kbytes']"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        ratio = self.json_data['jobs'][0]['read']['io_kbytes'] \
            / self.json_data['jobs'][0]['write']['io_kbytes']
        logging.debug("Test %d: ratio: %f", self.testnum, ratio)
        if ratio < 1.99 or ratio > 2.01:
            self.failure_reason = f"{self.failure_reason} read/write ratio mismatch,"
            self.passed = False


class FioJobFileTest_t0007(FioJobFileTest):
    """Test consists of fio test job t0007
    Confirm that read['io_kbytes'] = 87040"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        if self.json_data['jobs'][0]['read']['io_kbytes'] != 87040:
            self.failure_reason = f"{self.failure_reason} bytes read mismatch,"
            self.passed = False


class FioJobFileTest_t0008(FioJobFileTest):
    """Test consists of fio test job t0008
    Confirm that read['io_kbytes'] = 32768 and that
                write['io_kbytes'] ~ 16384

    This is a 50/50 seq read/write workload. Since fio flips a coin to
    determine whether to issue a read or a write, total bytes written will not
    be exactly 16384K. But total bytes read will be exactly 32768K because
    reads will include the initial phase as well as the verify phase where all
    the blocks originally written will be read."""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        ratio = self.json_data['jobs'][0]['write']['io_kbytes'] / 16384
        logging.debug("Test %d: ratio: %f", self.testnum, ratio)

        if ratio < 0.97 or ratio > 1.03:
            self.failure_reason = f"{self.failure_reason} bytes written mismatch,"
            self.passed = False
        if self.json_data['jobs'][0]['read']['io_kbytes'] != 32768:
            self.failure_reason = f"{self.failure_reason} bytes read mismatch,"
            self.passed = False


class FioJobFileTest_t0009(FioJobFileTest):
    """Test consists of fio test job t0009
    Confirm that runtime >= 60s"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        logging.debug('Test %d: elapsed: %d', self.testnum, self.json_data['jobs'][0]['elapsed'])

        if self.json_data['jobs'][0]['elapsed'] < 60:
            self.failure_reason = f"{self.failure_reason} elapsed time mismatch,"
            self.passed = False


class FioJobFileTest_t0012(FioJobFileTest):
    """Test consists of fio test job t0012
    Confirm ratios of job iops are 1:5:10
    job1,job2,job3 respectively"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        iops_files = []
        for i in range(1, 4):
            filename = os.path.join(self.paths['test_dir'], "{0}_iops.{1}.log".format(os.path.basename(
                self.fio_job), i))
            file_data = self.get_file_fail(filename)
            if not file_data:
                return

            iops_files.append(file_data.splitlines())

        # there are 9 samples for job1 and job2, 4 samples for job3
        iops1 = 0.0
        iops2 = 0.0
        iops3 = 0.0
        for i in range(9):
            iops1 = iops1 + float(iops_files[0][i].split(',')[1])
            iops2 = iops2 + float(iops_files[1][i].split(',')[1])
            iops3 = iops3 + float(iops_files[2][i].split(',')[1])

            ratio1 = iops3/iops2
            ratio2 = iops3/iops1
            logging.debug("sample {0}: job1 iops={1} job2 iops={2} job3 iops={3} " \
                "job3/job2={4:.3f} job3/job1={5:.3f}".format(i, iops1, iops2, iops3, ratio1,
                                                             ratio2))

        # test job1 and job2 succeeded to recalibrate
        if ratio1 < 1 or ratio1 > 3 or ratio2 < 7 or ratio2 > 13:
            self.failure_reason += " iops ratio mismatch iops1={0} iops2={1} iops3={2} " \
                "expected r1~2 r2~10 got r1={3:.3f} r2={4:.3f},".format(iops1, iops2, iops3,
                                                                        ratio1, ratio2)
            self.passed = False
            return


class FioJobFileTest_t0014(FioJobFileTest):
    """Test consists of fio test job t0014
	Confirm that job1_iops / job2_iops ~ 1:2 for entire duration
	and that job1_iops / job3_iops ~ 1:3 for first half of duration.

    The test is about making sure the flow feature can
    re-calibrate the activity dynamically"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        iops_files = []
        for i in range(1, 4):
            filename = os.path.join(self.paths['test_dir'], "{0}_iops.{1}.log".format(os.path.basename(
                self.fio_job), i))
            file_data = self.get_file_fail(filename)
            if not file_data:
                return

            iops_files.append(file_data.splitlines())

        # there are 9 samples for job1 and job2, 4 samples for job3
        iops1 = 0.0
        iops2 = 0.0
        iops3 = 0.0
        for i in range(9):
            if i < 4:
                iops3 = iops3 + float(iops_files[2][i].split(',')[1])
            elif i == 4:
                ratio1 = iops1 / iops2
                ratio2 = iops1 / iops3


                if ratio1 < 0.43 or ratio1 > 0.57 or ratio2 < 0.21 or ratio2 > 0.45:
                    self.failure_reason += " iops ratio mismatch iops1={0} iops2={1} iops3={2} " \
                                           "expected r1~0.5 r2~0.33 got r1={3:.3f} r2={4:.3f},".format(
                                               iops1, iops2, iops3, ratio1, ratio2)
                    self.passed = False

            iops1 = iops1 + float(iops_files[0][i].split(',')[1])
            iops2 = iops2 + float(iops_files[1][i].split(',')[1])

            ratio1 = iops1/iops2
            ratio2 = iops1/iops3
            logging.debug("sample {0}: job1 iops={1} job2 iops={2} job3 iops={3} " \
                          "job1/job2={4:.3f} job1/job3={5:.3f}".format(i, iops1, iops2, iops3,
                                                                       ratio1, ratio2))

        # test job1 and job2 succeeded to recalibrate
        if ratio1 < 0.43 or ratio1 > 0.57:
            self.failure_reason += " iops ratio mismatch iops1={0} iops2={1} expected ratio~0.5 " \
                                   "got ratio={2:.3f},".format(iops1, iops2, ratio1)
            self.passed = False
            return


class FioJobFileTest_t0015(FioJobFileTest):
    """Test consists of fio test jobs t0015 and t0016
    Confirm that mean(slat) + mean(clat) = mean(tlat)"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        slat = self.json_data['jobs'][0]['read']['slat_ns']['mean']
        clat = self.json_data['jobs'][0]['read']['clat_ns']['mean']
        tlat = self.json_data['jobs'][0]['read']['lat_ns']['mean']
        logging.debug('Test %d: slat %f, clat %f, tlat %f', self.testnum, slat, clat, tlat)

        if abs(slat + clat - tlat) > 1:
            self.failure_reason = "{0} slat {1} + clat {2} = {3} != tlat {4},".format(
                self.failure_reason, slat, clat, slat+clat, tlat)
            self.passed = False


class FioJobFileTest_t0019(FioJobFileTest):
    """Test consists of fio test job t0019
    Confirm that all offsets were touched sequentially"""

    def check_result(self):
        super().check_result()

        bw_log_filename = os.path.join(self.paths['test_dir'], "test_bw.log")
        file_data = self.get_file_fail(bw_log_filename)
        if not file_data:
            return

        log_lines = file_data.split('\n')

        prev = -4096
        for line in log_lines:
            if len(line.strip()) == 0:
                continue
            cur = int(line.split(',')[4])
            if cur - prev != 4096:
                self.passed = False
                self.failure_reason = f"offsets {prev}, {cur} not sequential"
                return
            prev = cur

        if cur/4096 != 255:
            self.passed = False
            self.failure_reason = f"unexpected last offset {cur}"


class FioJobFileTest_t0020(FioJobFileTest):
    """Test consists of fio test jobs t0020 and t0021
    Confirm that almost all offsets were touched non-sequentially"""

    def check_result(self):
        super().check_result()

        bw_log_filename = os.path.join(self.paths['test_dir'], "test_bw.log")
        file_data = self.get_file_fail(bw_log_filename)
        if not file_data:
            return

        log_lines = file_data.split('\n')

        offsets = []

        prev = int(log_lines[0].split(',')[4])
        for line in log_lines[1:]:
            offsets.append(prev/4096)
            if len(line.strip()) == 0:
                continue
            cur = int(line.split(',')[4])
            prev = cur

        if len(offsets) != 256:
            self.passed = False
            self.failure_reason += f" number of offsets is {len(offsets)} instead of 256"

        for i in range(256):
            if not i in offsets:
                self.passed = False
                self.failure_reason += f" missing offset {i * 4096}"

        (_, p) = runstest_1samp(list(offsets))
        if p < 0.05:
            self.passed = False
            self.failure_reason += f" runs test failed with p = {p}"


class FioJobFileTest_t0022(FioJobFileTest):
    """Test consists of fio test job t0022"""

    def check_result(self):
        super().check_result()

        bw_log_filename = os.path.join(self.paths['test_dir'], "test_bw.log")
        file_data = self.get_file_fail(bw_log_filename)
        if not file_data:
            return

        log_lines = file_data.split('\n')

        filesize = 1024*1024
        bs = 4096
        seq_count = 0
        offsets = set()

        prev = int(log_lines[0].split(',')[4])
        for line in log_lines[1:]:
            offsets.add(prev/bs)
            if len(line.strip()) == 0:
                continue
            cur = int(line.split(',')[4])
            if cur - prev == bs:
                seq_count += 1
            prev = cur

        # 10 is an arbitrary threshold
        if seq_count > 10:
            self.passed = False
            self.failure_reason = f"too many ({seq_count}) consecutive offsets"

        if len(offsets) == filesize/bs:
            self.passed = False
            self.failure_reason += " no duplicate offsets found with norandommap=1"


class FioJobFileTest_t0023(FioJobFileTest):
    """Test consists of fio test job t0023 randtrimwrite test."""

    def check_trimwrite(self, filename):
        """Make sure that trims are followed by writes of the same size at the same offset."""

        bw_log_filename = os.path.join(self.paths['test_dir'], filename)
        file_data = self.get_file_fail(bw_log_filename)
        if not file_data:
            return

        log_lines = file_data.split('\n')

        prev_ddir = 1
        for line in log_lines:
            if len(line.strip()) == 0:
                continue
            vals = line.split(',')
            ddir = int(vals[2])
            bs = int(vals[3])
            offset = int(vals[4])
            if prev_ddir == 1:
                if ddir != 2:
                    self.passed = False
                    self.failure_reason += " {0}: write not preceeded by trim: {1}".format(
                        bw_log_filename, line)
                    break
            else:
                if ddir != 1:   # pylint: disable=no-else-break
                    self.passed = False
                    self.failure_reason += " {0}: trim not preceeded by write: {1}".format(
                        bw_log_filename, line)
                    break
                else:
                    if prev_bs != bs:
                        self.passed = False
                        self.failure_reason += " {0}: block size does not match: {1}".format(
                            bw_log_filename, line)
                        break

                    if prev_offset != offset:
                        self.passed = False
                        self.failure_reason += " {0}: offset does not match: {1}".format(
                            bw_log_filename, line)
                        break

            prev_ddir = ddir
            prev_bs = bs
            prev_offset = offset


    def check_all_offsets(self, filename, sectorsize, filesize):
        """Make sure all offsets were touched."""

        file_data = self.get_file_fail(os.path.join(self.paths['test_dir'], filename))
        if not file_data:
            return

        log_lines = file_data.split('\n')

        offsets = set()

        for line in log_lines:
            if len(line.strip()) == 0:
                continue
            vals = line.split(',')
            bs = int(vals[3])
            offset = int(vals[4])
            if offset % sectorsize != 0:
                self.passed = False
                self.failure_reason += " {0}: offset {1} not a multiple of sector size {2}".format(
                    filename, offset, sectorsize)
                break
            if bs % sectorsize != 0:
                self.passed = False
                self.failure_reason += " {0}: block size {1} not a multiple of sector size " \
                    "{2}".format(filename, bs, sectorsize)
                break
            for i in range(int(bs/sectorsize)):
                offsets.add(offset/sectorsize + i)

        if len(offsets) != filesize/sectorsize:
            self.passed = False
            self.failure_reason += " {0}: only {1} offsets touched; expected {2}".format(
                filename, len(offsets), filesize/sectorsize)
        else:
            logging.debug("%s: %d sectors touched", filename, len(offsets))


    def check_result(self):
        super().check_result()

        filesize = 1024*1024

        self.check_trimwrite("basic_bw.log")
        self.check_trimwrite("bs_bw.log")
        self.check_trimwrite("bsrange_bw.log")
        self.check_trimwrite("bssplit_bw.log")
        self.check_trimwrite("basic_no_rm_bw.log")
        self.check_trimwrite("bs_no_rm_bw.log")
        self.check_trimwrite("bsrange_no_rm_bw.log")
        self.check_trimwrite("bssplit_no_rm_bw.log")

        self.check_all_offsets("basic_bw.log", 4096, filesize)
        self.check_all_offsets("bs_bw.log", 8192, filesize)
        self.check_all_offsets("bsrange_bw.log", 512, filesize)
        self.check_all_offsets("bssplit_bw.log", 512, filesize)


class FioJobFileTest_t0024(FioJobFileTest_t0023):
    """Test consists of fio test job t0024 trimwrite test."""

    def check_result(self):
        # call FioJobFileTest_t0023's parent to skip checks done by t0023
        super(FioJobFileTest_t0023, self).check_result()

        filesize = 1024*1024

        self.check_trimwrite("basic_bw.log")
        self.check_trimwrite("bs_bw.log")
        self.check_trimwrite("bsrange_bw.log")
        self.check_trimwrite("bssplit_bw.log")

        self.check_all_offsets("basic_bw.log", 4096, filesize)
        self.check_all_offsets("bs_bw.log", 8192, filesize)
        self.check_all_offsets("bsrange_bw.log", 512, filesize)
        self.check_all_offsets("bssplit_bw.log", 512, filesize)


class FioJobFileTest_t0025(FioJobFileTest):
    """Test experimental verify read backs written data pattern."""
    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        if self.json_data['jobs'][0]['read']['io_kbytes'] != 128:
            self.passed = False

class FioJobFileTest_t0027(FioJobFileTest):
    def setup(self, *args, **kws):
        super().setup(*args, **kws)
        self.pattern_file = os.path.join(self.paths['test_dir'], "t0027.pattern")
        self.output_file = os.path.join(self.paths['test_dir'], "t0027file")
        self.pattern = os.urandom(16 << 10)
        with open(self.pattern_file, "wb") as f:
            f.write(self.pattern)

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        with open(self.output_file, "rb") as f:
            data = f.read()

        if data != self.pattern:
            self.passed = False

class FioJobFileTest_t0029(FioJobFileTest):
    """Test loops option works with read-verify workload."""
    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        if self.json_data['jobs'][1]['read']['io_kbytes'] != 8:
            self.passed = False

class FioJobFileTest_LogFileFormat(FioJobFileTest):
    """Test log file format"""
    def setup(self, *args, **kws):
        super().setup(*args, **kws)
        self.patterns = {}

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        for logfile in self.patterns.keys():
            file_path = os.path.join(self.paths['test_dir'], logfile)
            with open(file_path, "r") as f:
                line = f.readline()
                if not re.match(self.patterns[logfile], line):
                    self.passed = False
                    self.failure_reason = "wrong log file format: " + logfile
                    return

class FioJobFileTest_t0033(FioJobFileTest_LogFileFormat):
    """Test log file format"""
    def setup(self, *args, **kws):
        super().setup(*args, **kws)
        self.patterns = {
            'log_bw.1.log': '\\d+, \\d+, \\d+, \\d+, 0x[\\da-f]+\\n',
            'log_clat.2.log': '\\d+, \\d+, \\d+, \\d+, 0, \\d+\\n',
            'log_iops.3.log': '\\d+, \\d+, \\d+, \\d+, \\d+, 0x[\\da-f]+\\n',
            'log_iops.4.log': '\\d+, \\d+, \\d+, \\d+, 0, 0, \\d+\\n',
        }

class FioJobFileTest_t0034(FioJobFileTest_LogFileFormat):
    """Test log file format"""
    def setup(self, *args, **kws):
        super().setup(*args, **kws)
        self.patterns = {
            'log_clat.1.log': '\\d+, \\d+, \\d+, \\d+, \\d+, \\d+, \\d+\\n',
            'log_slat.1.log': '\\d+, \\d+, \\d+, \\d+, \\d+, \\d+, \\d+\\n',
            'log_lat.1.log': '\\d+, \\d+, \\d+, \\d+, \\d+, \\d+, 0\\n',
            'log_clat.2.log': '\\d+, \\d+, \\d+, \\d+, 0, 0, \\d+, 0\\n',
            'log_bw.3.log': '\\d+, \\d+, \\d+, \\d+, \\d+, \\d+, 0\\n',
            'log_iops.3.log': '\\d+, \\d+, \\d+, \\d+, \\d+, \\d+, 0\\n',
        }

class FioJobFileTest_iops_rate(FioJobFileTest):
    """Test consists of fio test job t0011
    Confirm that job0 iops == 1000
    and that job1_iops / job0_iops ~ 8
    With two runs of fio-3.16 I observed a ratio of 8.3"""

    def check_result(self):
        super().check_result()

        if not self.passed:
            return

        iops1 = self.json_data['jobs'][0]['read']['iops']
        logging.debug("Test %d: iops1: %f", self.testnum, iops1)
        iops2 = self.json_data['jobs'][1]['read']['iops']
        logging.debug("Test %d: iops2: %f", self.testnum, iops2)
        ratio = iops2 / iops1
        logging.debug("Test %d: ratio: %f", self.testnum, ratio)

        if iops1 < 950 or iops1 > 1050:
            self.failure_reason = f"{self.failure_reason} iops value mismatch,"
            self.passed = False

        if ratio < 6 or ratio > 10:
            self.failure_reason = f"{self.failure_reason} iops ratio mismatch,"
            self.passed = False


TEST_LIST = [
    {
        'test_id':          1,
        'test_class':       FioJobFileTest,
        'job':              't0001-52c58027.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          2,
        'test_class':       FioJobFileTest,
        'job':              't0002-13af05ae-post.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          't0002-13af05ae-pre.fio',
        'pre_success':      None,
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          3,
        'test_class':       FioJobFileTest,
        'job':              't0003-0ae2c6e1-post.fio',
        'success':          SUCCESS_NONZERO,
        'pre_job':          't0003-0ae2c6e1-pre.fio',
        'pre_success':      SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          4,
        'test_class':       FioJobFileTest,
        'job':              't0004-8a99fdf6.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          5,
        'test_class':       FioJobFileTest_t0005,
        'job':              't0005-f7078f7b.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [Requirements.not_windows],
    },
    {
        'test_id':          6,
        'test_class':       FioJobFileTest_t0006,
        'job':              't0006-82af2a7c.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          7,
        'test_class':       FioJobFileTest_t0007,
        'job':              't0007-37cf9e3c.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          8,
        'test_class':       FioJobFileTest_t0008,
        'job':              't0008-ae2fafc8.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          9,
        'test_class':       FioJobFileTest_t0009,
        'job':              't0009-f8b0bd10.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [Requirements.not_macos,
                             Requirements.cpucount4],
        # mac os does not support CPU affinity
    },
    {
        'test_id':          10,
        'test_class':       FioJobFileTest,
        'job':              't0010-b7aae4ba.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          11,
        'test_class':       FioJobFileTest_iops_rate,
        'job':              't0011-5d2788d5.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          12,
        'test_class':       FioJobFileTest_t0012,
        'job':              't0012.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          13,
        'test_class':       FioJobFileTest,
        'job':              't0013.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          14,
        'test_class':       FioJobFileTest_t0014,
        'job':              't0014.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          15,
        'test_class':       FioJobFileTest_t0015,
        'job':              't0015-4e7e7898.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          16,
        'test_class':       FioJobFileTest_t0015,
        'job':              't0016-d54ae22.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          17,
        'test_class':       FioJobFileTest_t0015,
        'job':              't0017.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [Requirements.not_windows],
    },
    {
        'test_id':          18,
        'test_class':       FioJobFileTest,
        'job':              't0018.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [Requirements.linux, Requirements.io_uring],
    },
    {
        'test_id':          19,
        'test_class':       FioJobFileTest_t0019,
        'job':              't0019.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          20,
        'test_class':       FioJobFileTest_t0020,
        'job':              't0020.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          21,
        'test_class':       FioJobFileTest_t0020,
        'job':              't0021.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          22,
        'test_class':       FioJobFileTest_t0022,
        'job':              't0022.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          23,
        'test_class':       FioJobFileTest_t0023,
        'job':              't0023.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          24,
        'test_class':       FioJobFileTest_t0024,
        'job':              't0024.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          25,
        'test_class':       FioJobFileTest_t0025,
        'job':              't0025.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          26,
        'test_class':       FioJobFileTest,
        'job':              't0026.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [Requirements.not_windows],
    },
    {
        'test_id':          27,
        'test_class':       FioJobFileTest_t0027,
        'job':              't0027.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          28,
        'test_class':       FioJobFileTest,
        'job':              't0028-c6cade16.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'requirements':     [],
    },
    {
        'test_id':          29,
        'test_class':       FioJobFileTest_t0029,
        'job':              't0029.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'output_format':    'json',
        'requirements':     [],
    },
    {
        'test_id':          30,
        'test_class':       FioJobFileTest,
        'job':              't0030.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'parameters':       ['--bandwidth-log'],
        'requirements':     [],
    },
    {
        'test_id':          31,
        'test_class':       FioJobFileTest,
        'job':              't0031.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          't0031-pre.fio',
        'pre_success':      SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          33,
        'test_class':       FioJobFileTest_t0033,
        'job':              't0033.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'pre_success':      SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          34,
        'test_class':       FioJobFileTest_t0034,
        'job':              't0034.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'pre_success':      SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.libaio],
    },
    {
        'test_id':          35,
        'test_class':       FioJobFileTest,
        'job':              't0035.fio',
        'success':          SUCCESS_DEFAULT,
        'pre_job':          None,
        'pre_success':      None,
        'pre_success':      SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1000,
        'test_class':       FioExeTest,
        'exe':              't/axmap',
        'parameters':       None,
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1001,
        'test_class':       FioExeTest,
        'exe':              't/ieee754',
        'parameters':       None,
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1002,
        'test_class':       FioExeTest,
        'exe':              't/lfsr-test',
        'parameters':       ['0xFFFFFF', '0', '0', 'verify'],
        'success':          SUCCESS_STDERR,
        'requirements':     [],
    },
    {
        'test_id':          1003,
        'test_class':       FioExeTest,
        'exe':              't/readonly.py',
        'parameters':       ['-f', '{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1004,
        'test_class':       FioExeTest,
        'exe':              't/steadystate_tests.py',
        'parameters':       ['{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1005,
        'test_class':       FioExeTest,
        'exe':              't/stest',
        'parameters':       None,
        'success':          SUCCESS_STDERR,
        'requirements':     [],
    },
    {
        'test_id':          1006,
        'test_class':       FioExeTest,
        'exe':              't/strided.py',
        'parameters':       ['--fio', '{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1007,
        'test_class':       FioExeTest,
        'exe':              't/zbd/run-tests-against-nullb',
        'parameters':       ['-s', '1'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.zbd,
                             Requirements.root],
    },
    {
        'test_id':          1008,
        'test_class':       FioExeTest,
        'exe':              't/zbd/run-tests-against-nullb',
        'parameters':       ['-s', '2'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.zbd,
                             Requirements.root, Requirements.zoned_nullb],
    },
    {
        'test_id':          1009,
        'test_class':       FioExeTest,
        'exe':              'unittests/unittest',
        'parameters':       None,
        'success':          SUCCESS_DEFAULT,
        'requirements':     [Requirements.unittests],
    },
    {
        'test_id':          1010,
        'test_class':       FioExeTest,
        'exe':              't/latency_percentiles.py',
        'parameters':       ['-f', '{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1011,
        'test_class':       FioExeTest,
        'exe':              't/jsonplus2csv_test.py',
        'parameters':       ['-f', '{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1012,
        'test_class':       FioExeTest,
        'exe':              't/log_compression.py',
        'parameters':       ['-f', '{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1013,
        'test_class':       FioExeTest,
        'exe':              't/random_seed.py',
        'parameters':       ['-f', '{fio_path}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [],
    },
    {
        'test_id':          1014,
        'test_class':       FioExeTest,
        'exe':              't/nvmept.py',
        'parameters':       ['-f', '{fio_path}', '--dut', '{nvmecdev}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.nvmecdev],
    },
    {
        'test_id':          1015,
        'test_class':       FioExeTest,
        'exe':              't/nvmept_trim.py',
        'parameters':       ['-f', '{fio_path}', '--dut', '{nvmecdev}'],
        'success':          SUCCESS_DEFAULT,
        'requirements':     [Requirements.linux, Requirements.nvmecdev],
    },
]


def parse_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--fio-root',
                        help='fio root path')
    parser.add_argument('-f', '--fio',
                        help='path to fio executable (e.g., ./fio)')
    parser.add_argument('-a', '--artifact-root',
                        help='artifact root directory')
    parser.add_argument('-s', '--skip', nargs='+', type=int,
                        help='list of test(s) to skip')
    parser.add_argument('-o', '--run-only', nargs='+', type=int,
                        help='list of test(s) to run, skipping all others')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='provide debug output')
    parser.add_argument('-k', '--skip-req', action='store_true',
                        help='skip requirements checking')
    parser.add_argument('-p', '--pass-through', action='append',
                        help='pass-through an argument to an executable test')
    parser.add_argument('--nvmecdev', action='store', default=None,
                        help='NVMe character device for **DESTRUCTIVE** testing (e.g., /dev/ng0n1)')
    args = parser.parse_args()

    return args


def main():
    """Entry point."""

    args = parse_args()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    pass_through = {}
    if args.pass_through:
        for arg in args.pass_through:
            if not ':' in arg:
                print(f"Invalid --pass-through argument '{arg}'")
                print("Syntax for --pass-through is TESTNUMBER:ARGUMENT")
                return
            split = arg.split(":", 1)
            pass_through[int(split[0])] = split[1]
        logging.debug("Pass-through arguments: %s", pass_through)

    if args.fio_root:
        fio_root = args.fio_root
    else:
        fio_root = str(Path(__file__).absolute().parent.parent)
    print(f"fio root is {fio_root}")

    if args.fio:
        fio_path = args.fio
    else:
        if platform.system() == "Windows":
            fio_exe = "fio.exe"
        else:
            fio_exe = "fio"
        fio_path = os.path.join(fio_root, fio_exe)
    print(f"fio path is {fio_path}")
    if not shutil.which(fio_path):
        print("Warning: fio executable not found")

    artifact_root = args.artifact_root if args.artifact_root else \
        f"fio-test-{time.strftime('%Y%m%d-%H%M%S')}"
    os.mkdir(artifact_root)
    print(f"Artifact directory is {artifact_root}")

    if not args.skip_req:
        Requirements(fio_root, args)

    test_env = {
              'fio_path': fio_path,
              'fio_root': fio_root,
              'artifact_root': artifact_root,
              'pass_through': pass_through,
              }
    _, failed, _ = run_fio_tests(TEST_LIST, test_env, args)
    sys.exit(failed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# steadystate_tests.py
#
# Test option parsing and functionality for fio's steady state detection feature.
#
# steadystate_tests.py --read file-for-read-testing --write file-for-write-testing ./fio
#
# REQUIREMENTS
# Python 2.6+
# SciPy
#
# KNOWN ISSUES
# only option parsing and read tests are carried out
# On Windows this script works under Cygwin but not from cmd.exe
# On Windows I encounter frequent fio problems generating JSON output (nothing to decode)
# min runtime:
# if ss attained: min runtime = ss_dur + ss_ramp
# if not attained: runtime = timeout

from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import json
import pprint
import argparse
import subprocess
from scipy import stats

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('fio', help='path to fio executable')
    args = parser.parse_args()

    return args


def check(data, iops, slope, pct, limit, dur, criterion):
    measurement = 'iops' if iops else 'bw'
    data = data[measurement]
    mean = sum(data) / len(data)
    if slope:
        x = list(range(len(data)))
        m, intercept, r_value, p_value, std_err = stats.linregress(x,data)
        m = abs(m)
        if pct:
            target = (m / mean * 100) if mean != 0 else 0
            criterion = criterion[:-1]
        else:
            target = m
    else:
        maxdev = 0
        for x in data:
            maxdev = max(abs(mean-x), maxdev)
        if pct:
            target = maxdev / mean * 100
            criterion = criterion[:-1]
        else:
            target = maxdev

    criterion = float(criterion)
    if criterion == 0.0:
        objsame = False
    else:
        objsame = abs(target - criterion) / criterion < 0.005
    return (objsame, target < limit, mean, target)


if __name__ == '__main__':
    args = parse_args()

    pp = pprint.PrettyPrinter(indent=4)

    passed = 0
    failed = 0

#
# test option parsing
#
    parsing = [ { 'args': ["--parse-only", "--debug=parse", "--ss_dur=10s", "--ss=iops:10", "--ss_ramp=5"],
                  'output': "set steady state IOPS threshold to 10.000000" },
                { 'args': ["--parse-only", "--debug=parse", "--ss_dur=10s", "--ss=iops:10%", "--ss_ramp=5"],
                  'output': "set steady state threshold to 10.000000%" },
                { 'args': ["--parse-only", "--debug=parse", "--ss_dur=10s", "--ss=iops:.1%", "--ss_ramp=5"],
                  'output': "set steady state threshold to 0.100000%" },
                { 'args': ["--parse-only", "--debug=parse", "--ss_dur=10s", "--ss=bw:10%", "--ss_ramp=5"],
                  'output': "set steady state threshold to 10.000000%" },
                { 'args': ["--parse-only", "--debug=parse", "--ss_dur=10s", "--ss=bw:.1%", "--ss_ramp=5"],
                  'output': "set steady state threshold to 0.100000%" },
                { 'args': ["--parse-only", "--debug=parse", "--ss_dur=10s", "--ss=bw:12", "--ss_ramp=5"],
                  'output': "set steady state BW threshold to 12" },
              ]
    for test in parsing:
        output = subprocess.check_output([args.fio] + test['args'])
        if test['output'] in output.decode():
            print("PASSED '{0}' found with arguments {1}".format(test['output'], test['args']))
            passed = passed + 1
        else:
            print("FAILED '{0}' NOT found with arguments {1}".format(test['output'], test['args']))
            failed = failed + 1

#
# test some read workloads
#
# if ss active and attained,
#   check that runtime is less than job time
#   check criteria
#   how to check ramp time?
#
# if ss inactive
#   check that runtime is what was specified
#
    reads = [ {'s': True, 'timeout': 100, 'numjobs': 1, 'ss_dur': 5, 'ss_ramp': 3, 'iops': True, 'slope': True, 'ss_limit': 0.1, 'pct': True},
              {'s': False, 'timeout': 20, 'numjobs': 2},
              {'s': True, 'timeout': 100, 'numjobs': 3, 'ss_dur': 10, 'ss_ramp': 5, 'iops': False, 'slope': True, 'ss_limit': 0.1, 'pct': True},
              {'s': True, 'timeout': 10, 'numjobs': 3, 'ss_dur': 10, 'ss_ramp': 500, 'iops': False, 'slope': True, 'ss_limit': 0.1, 'pct': True},
              {'s': True, 'timeout': 10, 'numjobs': 3, 'ss_dur': 10, 'ss_ramp': 500, 'iops': False, 'slope': True, 'ss_limit': 0.1, 'pct': True, 'ss_interval': 5},
            ]

    jobnum = 0
    for job in reads:

        tf = "steadystate_job{0}.json".format(jobnum)
        parameters = [ "--max-jobs=16", "--name=job{0}".format(jobnum) ]
        parameters.extend([ "--thread",
                            "--output-format=json",
                            "--output={0}".format(tf),
                            "--ioengine=null",
                            "--size=1G",
                            "--rw=randrw",
                            "--rwmixread=100",
                            "--stonewall",
                            "--group_reporting",
                            "--numjobs={0}".format(job['numjobs']),
                            "--time_based",
                            "--runtime={0}".format(job['timeout']) ])
        if job['s']:
           if job['iops']:
               ss = 'iops'
           else:
               ss = 'bw'
           if job['slope']:
               ss += "_slope"
           ss += ":" + str(job['ss_limit'])
           if job['pct']:
               ss += '%'
           parameters.extend([ '--ss_dur={0}'.format(job['ss_dur']),
                               '--ss={0}'.format(ss),
                               '--ss_ramp={0}'.format(job['ss_ramp']) ])

        output = subprocess.call([args.fio] + parameters)
        with open(tf, 'r') as source:
            jsondata = json.loads(source.read())
            source.close()

        for jsonjob in jsondata['jobs']:
            line = "{0}".format(jsonjob['job options']['name'])
            if job['s']:
                if jsonjob['steadystate']['attained'] == 1:
                    # check runtime >= ss_dur + ss_ramp, check criterion, check criterion < limit
                    mintime = (job['ss_dur'] + job['ss_ramp']) * 1000
                    actual = jsonjob['read']['runtime']
                    if mintime > actual:
                        line = 'FAILED ' + line + ' ss attained, runtime {0} < ss_dur {1} + ss_ramp {2}'.format(actual, job['ss_dur'], job['ss_ramp'])
                        failed = failed + 1
                    else:
                        line = line + ' ss attained, runtime {0} > ss_dur {1} + ss_ramp {2},'.format(actual, job['ss_dur'], job['ss_ramp'])
                        objsame, met, mean, target = check(data=jsonjob['steadystate']['data'],
                            iops=job['iops'],
                            slope=job['slope'],
                            pct=job['pct'],
                            limit=job['ss_limit'],
                            dur=job['ss_dur'],
                            criterion=jsonjob['steadystate']['criterion'])
                        if not objsame:
                            line = 'FAILED ' + line + ' fio criterion {0} != calculated criterion {1} '.format(jsonjob['steadystate']['criterion'], target)
                            failed = failed + 1
                        else:
                            if met:
                                line = 'PASSED ' + line + ' target {0} < limit {1}'.format(target, job['ss_limit'])
                                passed = passed + 1
                            else:
                                line = 'FAILED ' + line + ' target {0} < limit {1} but fio reports ss not attained '.format(target, job['ss_limit'])
                                failed = failed + 1
                else:
                    # check runtime, confirm criterion calculation, and confirm that criterion was not met
                    expected = job['timeout'] * 1000
                    actual = jsonjob['read']['runtime']
                    if abs(expected - actual) > 50:
                        line = 'FAILED ' + line + ' ss not attained, expected runtime {0} != actual runtime {1}'.format(expected, actual)
                    else:
                        line = line + ' ss not attained, runtime {0} != ss_dur {1} + ss_ramp {2},'.format(actual, job['ss_dur'], job['ss_ramp'])
                        objsame, met, mean, target = check(data=jsonjob['steadystate']['data'],
                            iops=job['iops'],
                            slope=job['slope'],
                            pct=job['pct'],
                            limit=job['ss_limit'],
                            dur=job['ss_dur'],
                            criterion=jsonjob['steadystate']['criterion'])
                        if not objsame:
                            if actual > (job['ss_dur'] + job['ss_ramp'])*1000:
                                line = 'FAILED ' + line + ' fio criterion {0} != calculated criterion {1} '.format(jsonjob['steadystate']['criterion'], target)
                                failed = failed + 1
                            else:
                                line = 'PASSED ' + line + ' fio criterion {0} == 0.0 since ss_dur + ss_ramp has not elapsed '.format(jsonjob['steadystate']['criterion'])
                                passed = passed + 1
                        else:
                            if met:
                                line = 'FAILED ' + line + ' target {0} < threshold {1} but fio reports ss not attained '.format(target, job['ss_limit'])
                                failed = failed + 1
                            else:
                                line = 'PASSED ' + line + ' criterion {0} > threshold {1}'.format(target, job['ss_limit'])
                                passed = passed + 1
            else:
                expected = job['timeout'] * 1000
                actual = jsonjob['read']['runtime']
                if abs(expected - actual) > 50:
                    result = 'FAILED '
                    failed = failed + 1
                else:
                    result = 'PASSED '
                    passed = passed + 1
                line = result + line + ' no ss, expected runtime {0} ~= actual runtime {1}'.format(expected, actual)
            print(line)
            if 'steadystate' in jsonjob:
                pp.pprint(jsonjob['steadystate'])
        jobnum += 1

    print("{0} test(s) PASSED, {1} test(s) FAILED".format(passed,failed))
    sys.exit(failed)



#!/usr/bin/env python3
import uuid
import time
import errno
from graphviz import Digraph
import argparse
import configparser
import os

config_file = None
fio_file = None


def get_section_option(section_name, option_name, default=None):
    global fio_file
    if fio_file.has_option(section_name, option_name):
        return fio_file[section_name][option_name]
    return default


def get_config_option(section_name, option_name, default=None):
    global config_file
    if config_file.has_option(section_name, option_name):
        return config_file[section_name][option_name]
    return default


def get_header_color(keyword='fio_jobs', default_color='black'):
    return get_config_option(keyword, 'header_color', default_color)


def get_shape_color(keyword='fio_jobs', default_color='black'):
    return get_config_option(keyword, 'shape_color', default_color)


def get_text_color(keyword='fio_jobs', default_color='black'):
    return get_config_option(keyword, 'text_color', default_color)


def get_cluster_color(keyword='fio_jobs', default_color='gray92'):
    return get_config_option(keyword, 'cluster_color', default_color)


def get_header(keyword='fio_jobs'):
    return get_config_option(keyword, 'header')


def get_shape(keyword='fio_jobs'):
    return get_config_option(keyword, 'shape', 'box')


def get_style(keyword='fio_jobs'):
    return get_config_option(keyword, 'style', 'rounded')


def get_cluster_style(keyword='fio_jobs'):
    return get_config_option(keyword, 'cluster_style', 'filled')


def get_specific_options(engine):
    if not engine:
        return ''
    return get_config_option('ioengine_{}'.format(engine), 'specific_options', '').split(' ')


def render_option(section, label, display, option, color_override=None):
    # These options are already shown with graphical helpers, no need to report them directly
    skip_list = ['size', 'stonewall', 'runtime', 'time_based',
                 'numjobs', 'wait_for', 'wait_for_previous']
    # If the option doesn't exist or if a special handling is already done
    # don't render it, just return the current state
    if option in skip_list or option not in section:
        return label, display
    display = option
    if section[option]:
        display = '{} = {}'.format(display, section[option])

    # Adding jobs's options into the box, darkgreen is the default color
    if color_override:
        color = color_override
    else:
        color = get_text_color(option, get_text_color('fio_jobs', 'darkgreen'))
    label += get_config_option('fio_jobs',
                               'item_style').format(color, display)
    return label, display


def render_options(fio_file, section_name):
    """Render all options of a section."""
    display = section_name
    section = fio_file[section_name]

    # Add a multiplier to the section_name if numjobs is set
    numjobs = int(get_section_option(section_name, 'numjobs', '1'))
    if numjobs > 1:
        display = display + \
            get_style('numjobs').format(
                get_text_color('numjobs'), numjobs)

    # Header of the box
    label = get_config_option('fio_jobs', 'title_style').format(display)

    # Let's parse all the options of the current fio thread
    # Some needs to be printed on top or bottom of the job to ease the read
    to_early_print = ['exec_prerun', 'ioengine']
    to_late_print = ['exec_postrun']

    # Let's print the options on top of the box
    for early_print in to_early_print:
        label, display = render_option(
            section, label, display, early_print)

    current_io_engine = get_section_option(
        section_name, 'ioengine', None)
    if current_io_engine:
        # Let's print all specifics options for this engine
        for specific_option in sorted(get_specific_options(current_io_engine)):
            label, display = render_option(
                section, label, display, specific_option, get_config_option('ioengine', 'specific_options_color'))

    # Let's print generic options sorted by name
    for option in sorted(section):
        if option in to_early_print or option in to_late_print or option in get_specific_options(current_io_engine):
            continue
        label, display = render_option(section, label, display, option)

    # let's print options on the bottom of the box
    for late_print in to_late_print:
        label, display = render_option(
            section, label, display, late_print)

    # End of the box content
    label += '</table>>'
    return label


def render_section(current_graph, fio_file, section_name, label):
    """Render the section."""
    attr = None
    section = fio_file[section_name]

    # Let's render the box associated to a job
    current_graph.node(section_name, label,
                       shape=get_shape(),
                       color=get_shape_color(),
                       style=get_style())

    # Let's report the duration of the jobs with a self-loop arrow
    if 'runtime' in section and 'time_based' in section:
        attr = 'runtime={}'.format(section['runtime'])
    elif 'size' in section:
        attr = 'size={}'.format(section['size'])
    if attr:
        current_graph.edge(section_name, section_name, attr)


def create_sub_graph(name):
    """Return a new graph."""
    # We need to put 'cluster' in the name to ensure graphviz consider it as a cluster
    cluster_name = 'cluster_' + name
    # Unset the main graph labels to avoid a recopy in each subgraph
    attr = {}
    attr['label'] = ''
    new_graph = Digraph(name=cluster_name, graph_attr=attr)
    new_graph.attr(style=get_cluster_style(),
                   color=get_cluster_color())
    return new_graph


def create_legend():
    """Return a legend."""
    html_table = "<<table border='0' cellborder='1' cellspacing='0' cellpadding='4'>"
    html_table += '<tr><td COLSPAN="2"><b>Legend</b></td></tr>'
    legend_item = '<tr> <td>{}</td> <td><font color="{}">{}</font></td></tr>"'
    legend_bgcolor_item = '<tr><td>{}</td><td BGCOLOR="{}"></td></tr>'
    html_table += legend_item.format('numjobs',
                                     get_text_color('numjobs'), 'x numjobs')
    html_table += legend_item.format('generic option',
                                     get_text_color(), 'generic option')
    html_table += legend_item.format('ioengine option',
                                     get_text_color('ioengine'), 'ioengine option')
    html_table += legend_bgcolor_item.format('job', get_shape_color())
    html_table += legend_bgcolor_item.format(
        'execution group', get_cluster_color())
    html_table += '</table>>'
    legend = Digraph('html_table')
    legend.node('legend', shape='none', label=html_table)
    return legend


def fio_to_graphviz(filename, format):
    """Compute the graphviz graph from the fio file."""

    # Let's read the fio file
    global fio_file
    fio_file = configparser.RawConfigParser(
        allow_no_value=True,
        default_section="global",
        inline_comment_prefixes="'#', ';'")
    fio_file.read(filename)

    # Prepare the main graph object
    # Let's define the header of the document
    attrs = {}
    attrs['labelloc'] = 't'
    attrs['label'] = get_header().format(
        get_header_color(), os.path.basename(filename))
    main_graph = Digraph(engine='dot', graph_attr=attrs, format=format)

    # Let's add a legend
    main_graph.subgraph(create_legend())

    # By default all jobs are run in parallel and depends on "global"
    depends_on = fio_file.default_section

    # The previous section is by default the global section
    previous_section = fio_file.default_section

    current_graph = main_graph

    # The first job will be a new execution group
    new_execution_group = True

    # Let's iterate on all sections to create links between them
    for section_name in fio_file.sections():
        # The current section
        section = fio_file[section_name]

        # If the current section is waiting the previous job
        if ('stonewall' or 'wait_for_previous') in section:
            # let's remember what was the previous job we depend on
            depends_on = previous_section
            new_execution_group = True
        elif 'wait_for' in section:
            # This sections depends on a named section pointed by wait_for
            depends_on = section['wait_for']
            new_execution_group = True

        if new_execution_group:
            # Let's link the current graph with the main one
            main_graph.subgraph(current_graph)
            # Let's create a new graph to represent all the incoming jobs running at the same time
            current_graph = create_sub_graph(section_name)

        # Let's render the current section in its execution group
        render_section(current_graph, fio_file, section_name,
                       render_options(fio_file, section_name))

        # Let's trace the link between this job and the one it depends on
        # If we depend on 'global', we can avoid doing adding an arrow as we don't want to see 'global'
        if depends_on != fio_file.default_section:
            current_graph.edge(depends_on, section_name)

        # The current section become the parent of the next one
        previous_section = section_name

        # We are by default in the same execution group
        new_execution_group = False

    # The last subgraph isn't rendered yet
    main_graph.subgraph(current_graph)

    # Let's return the main graphviz object
    return main_graph


def setup_commandline():
    "Prepare the command line."
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', action='store',
                        type=str,
                        required=True,
                        help='the fio file to graph')
    parser.add_argument('--output', action='store',
                        type=str,
                        help='the output filename')
    parser.add_argument('--format', action='store',
                        type=str,
                        default='png',
                        help='the output format (see https://graphviz.org/docs/outputs/)')
    parser.add_argument('--view', action='store_true',
                        default=False,
                        help='view the graph')
    parser.add_argument('--keep', action='store_true',
                        default=False,
                        help='keep the graphviz script file')
    parser.add_argument('--config', action='store',
                        type=str,
                        help='the configuration filename')
    args = parser.parse_args()
    return args


def main():
    global config_file
    args = setup_commandline()

    if args.config is None:
        if os.path.exists('fiograph.conf'):
            config_filename = 'fiograph.conf'
        else:
            config_filename = os.path.join(os.path.dirname(__file__), 'fiograph.conf')
            if not os.path.exists(config_filename):
                raise FileNotFoundError("Cannot locate configuration file")
    else:
        config_filename = args.config
    config_file = configparser.RawConfigParser(allow_no_value=True)
    config_file.read(config_filename)

    temp_filename = uuid.uuid4().hex
    image_filename = fio_to_graphviz(args.file, args.format).render(temp_filename, view=args.view)

    output_filename_stub = args.file
    if args.output:
        output_filename = args.output
    else:
        if output_filename_stub.endswith('.fio'):
            output_filename_stub = output_filename_stub[:-4]
        output_filename = image_filename.replace(temp_filename, output_filename_stub)
    if args.view:
        time.sleep(1)
        # allow time for the file to be opened before renaming it
    os.rename(image_filename, output_filename)

    if not args.keep:
        os.remove(temp_filename)
    else:
        os.rename(temp_filename, output_filename_stub + '.gv')


main()
